
#include "UmlItem.h"
#include "UmlDiagram.h"
#include "UmlTypeSpec.h"

#include <qdir.h>
#include <qfiledialog.h>
#include <qmessagebox.h>

#include "UmlClass.h"
#include "UmlOperation.h"
#include "UmlClassDiagram.h"
#include "UmlStateDiagram.h"
#include "UmlUseCaseDiagram.h"
#include "UmlSequenceDiagram.h"
#include "UmlCollaborationDiagram.h"
#include "UmlComponentDiagram.h"
#include "UmlDeploymentDiagram.h"
#include "UmlBasePackage.h"
#include "UmlCom.h"
#include "CppSettings.h"
#include "JavaSettings.h"

 UmlItem::~UmlItem() {
}

QCString UmlItem::sKind() {
  return "???";
}

void UmlItem::set_dir() {
  if (!UmlBasePackage::getProject()->propertyValue("html dir", directory))
    directory = "/tmp/" + name() + "_html";

  QString d(directory);

  d = QFileDialog::getExistingDirectory(d, 0, "","Directory where the files will be produced", TRUE);
  
  if (d.isEmpty())
    throw 0;
    
  directory = (const char *) d;
  
  QDir dir(d);
  
  if (dir.exists()) {
    bool rem = FALSE;
    
    // remove old html files ?
    for (unsigned i = 0; i != dir.count(); i += 1) {
      if (dir[i].right(5).lower() == ".html") {
	if (!rem &&
	    (QMessageBox::critical(0, "Html generator", 
				   "Delete already existing html files ?",
				   "Yes", "No", QString::null, 
				   1, 1) != 0))
	  break;
	rem = TRUE;
	dir.remove(dir[i]);
      }
    }
  }
  else if (! dir.mkdir(d)) {
    UmlCom::trace("Cannot create directory " + directory);
    throw 0;
  }

  UmlBasePackage::getProject()->set_PropertyValue("html dir", directory);
}

void UmlItem::memo_ref() {
  all.addElement(this);
  
  refId = ++nrefs;
  
  const QVector<UmlItem> ch = children();
  
  for (unsigned i = 0; i != ch.size(); i += 1)
    ch[i]->memo_ref();
}

void UmlItem::define() {
  fw.write("<A NAME=\"ref");
  fw.write(refId);
  fw.write("\"></a>\n");
}

void UmlItem::start_file(QCString f, QCString s, bool withrefs, UmlItem * x)
{
  QCString filename = directory + f;
  
  if (! fw.open(filename + ".html"))
    throw 0;
  
  fw.write("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">\n");
  fw.write("<HTML>\n");
  fw.write("\n");
  fw.write("<HEAD>\n");
  
  if (s.isEmpty()) {
    fw.write("<TITLE>full document</TITLE>\n");
    fw.write("</HEAD>\n");
  }
  else {
    fw.write("<TITLE>"); writeq(s); fw.write("</TITLE>\n");
    fw.write("</HEAD>\n");
    fw.write("\n");
    fw.write("<BODY bgcolor=\"#ffffff\">\n");
    fw.write("\n");
    if (x != 0)
      x->define();
    if (withrefs)
      ref_indexes();
    fw.write("<TABLE BORDER=\"0\" WIDTH=\"100%\" BGCOLOR=\"#87ceff\">\n");
    fw.write("\n");  fw.write("<TD ALIGN=CENTER>\n");
    fw.write("<font size=\"+7\"><b>"); writeq(s); fw.write("</b></font>\n");
    fw.write("<p>\n");
    fw.write("</TD>\n");
    fw.write("\n");
    fw.write("</TABLE>\n");
    fw.write("<p>\n");
    fw.write("\n");
    fw.write("<!-- ============================================================= -->\n");
    fw.write("\n");
  }
}

void UmlItem::end_file()
{
  fw.write("</BODY>\n");
  fw.write("</HTML>\n");
    
  fw.close();

}

void UmlItem::ref_indexes()
{
  fw.write("<p><hr noshade></p>\n<p>");
  fw.write("<a href=\"index.html\"><b> -Top- </b></a>");
  
  UmlClass::ref_index();
  UmlOperation::ref_index();
  UmlClassDiagram::ref_index();
  UmlStateDiagram::ref_index();
  UmlUseCaseDiagram::ref_index();
  UmlSequenceDiagram::ref_index();
  UmlCollaborationDiagram::ref_index();
  UmlComponentDiagram::ref_index();
  UmlDeploymentDiagram::ref_index();
  fw.write("</p>\n<p>\n</p>\n<p>");
  
  for (unsigned i = 0; i != letters.length(); i += 1) {
    fw.write("<a href=\"index_");
    fw.write(letters.at(i));
    fw.write(".html\"><b> ");
    fw.write(letters.at(i));
    fw.write(" </b></a>");
  }
  
  fw.write("</p>\n<p><hr noshade></p>\n");
}

void UmlItem::generate_indexes()
{
  UmlClass::generate_index();
  UmlOperation::generate_index();
  UmlClassDiagram::generate_index();
  UmlStateDiagram::generate_index();
  UmlUseCaseDiagram::generate_index();
  UmlSequenceDiagram::generate_index();
  UmlCollaborationDiagram::generate_index();
  UmlComponentDiagram::generate_index();
  UmlDeploymentDiagram::generate_index();

  int n = all.size();
  int i;
  char previous;
  
  sort(all);
  
  previous = 0;
  for (i = 0; i != n; i += 1) {
    UmlItem * x = all.elementAt(i);
    
    if (x->refId != 0) {
      QCString s = x->pretty_name();
      
      if (! s.isEmpty()) {
	char c = *((const char *) s);
	
	if ((c >= 'a') && (c <= 'z'))
	  c += 'A' - 'a';
	
	if (c != previous) {
	  previous = c;
	  letters += c;
	}
      }
    }
  }

  previous = 0;
  for (i = 0; i != n; i += 1) {
    UmlItem * x = all.elementAt(i);
    
    if (x->refId != 0) {
      QCString s = x->pretty_name();
      
      if (! s.isEmpty()) {
	char c = *((const char *) s);
	
	if ((c >= 'a') && (c <= 'z'))
	  c += 'A' - 'a';
	
	if (c != previous) {
	  if (previous != 0) {
	    fw.write("</table>\n");
	    end_file();
	  }
	  
	  previous = c;
	  
	  start_file(QCString("index_") + c, QCString("") + c, TRUE, 0);
	  
	  fw.write("<table>\n");
	  fw.write("<tr bgcolor=#f0f0f0><td align=center><b>Name</b></td><td align=center><b>Kind</b></td><td align=center><b>Description</b></td></tr>\n");
	}
	
	fw.write("<tr bgcolor=#f0f0f0><td>");
	x->write();
	fw.write("</td><td>");
	fw.write(x->sKind());
	fw.write("</td><td>");
	writeq(x->description());
	fw.write("</td></tr>\n");
      }
    }
  }

  if (previous != 0) {
    fw.write("</table>\n");
    end_file();
  }
}

void UmlItem::frame()
{
  start_file("index-withframe", "", FALSE, 0);

  fw.write("<FRAMESET cols=\"20%,80%\">\n");
  fw.write("<FRAME src=\"classes_list.html\" name=\"classesListFrame\">");
  fw.write("<FRAME src=\"index.html\" name=\"projectFrame\">");
  fw.write("</FRAMESET>");
  fw.write("<NOFRAMES>");
  fw.write("<H2>");
  fw.write("Frame Alert</H2>");
  fw.write("<P>");
  fw.write("This document is designed to be viewed using the frames feature. If you see this message, you are using a non-frame-capable web client.");
  fw.write("<BR>");
  fw.write("Link to <A HREF=\"index.html\">Non-frame version.</A></NOFRAMES>");
  
  end_file();

  UmlCom::trace("document with frame produced in <i>"
		+ directory + "index-withframe.html");

  UmlCom::trace("document without frame produced in <i>"
		+ directory + "index.html");
}

void UmlItem::html(QCString pfix, unsigned int rank, QCString what) {
  define();
 
  chapter(what, pfix, rank, "98fb98");

  if (! description().isEmpty()) {
    fw.write("<p>");
    writeq(description());
    fw.write("<br></p>\n");
  }
  
  const QVector<UmlItem> ch = children();
  unsigned n = ch.size();
  
  if (n != 0) {
    QCString spfix;
    
    if (rank != 0) {
      spfix.setNum(rank);
      spfix = pfix + spfix + ".";
    }
    
    for (unsigned i = 0; i != n; i += 1)
      ch[i]->html(spfix, i + 1);
  }
}

void UmlItem::html(const char * what, UmlDiagram * diagram) {
  define();

  fw.write("<table bgcolor=\"#d3d3d3\"><td><font size=\"+1\">");
  fw.write(what);
  fw.write("  <b>");
  writeq(name());
  fw.write("</b></font><br></td></table>\n");

  QCString d = description();
  
  if (!d.isEmpty()) {
    fw.write("<p>");
    writeq(d);
    fw.write("<br></p>\n");
  }
  
  if (diagram != 0) {
    fw.write("<p>Diagram : ");
    diagram->write();
    fw.write("</p>\n");
  }
}

void UmlItem::chapter(QCString k, QCString pfix, unsigned int rank, QCString color) {
  if (rank != 0) {
    fw.write("<table bgcolor=\"#");
    fw.write(color);
    fw.write("\"><td><font size=\"+2\">");
    fw.write(pfix);
    fw.write(rank);
    fw.write(' ');
    fw.write(k);
    fw.write(" <b>");
    writeq(name());
    fw.write("</b></font><br></td></table>\n");
  }
}

const char * UmlItem::bypass_comment(const char * s) {
  while (*s) {
    if (*s <= ' ')
      s += 1;
    else if ((*s == '/') && (s[1] != 0)) {
      if (s[1] == '/') {
	do 
	  s += 1;
	while ((*s != 0) &&
	       (*s != '\n') && (*s != '\r'));
      }
      else if (s[1] == '*') {
	const char * s2 = strstr(s + 2, "*/");
	
	if (s2 != 0)
	  s = s2 + 2;
	else
	  return s;
      }
      else
	return s;
    }
    else if (*s == '#') {
      do {
	if ((*++s == 0) ||
	    ((*s == '\\') && (*++s == 0)))
	  return s;
      } while ((*s != '\n') && (*s != '\r'));
    }
    else
      return s;
  }

  return s;
}

void UmlItem::manage_alias(const char *& p) {
  // p starts by '@'
  const char * pclosed;
  
  if ((p[1] == '{') && ((pclosed = strchr(p + 2, '}')) != 0)) {
    QCString key(p + 2, pclosed - p - 1);
    QCString value;
    UmlItem * node = this;

    do {
      if (node->propertyValue(key, value))
	break;
      node = node->parent();
    } while (node != 0);
    
    if (node != 0)
      // find, insert the value
      writeq(value);
    else {
      // not find, insert the key
      fw.write("@{");
      writeq(key);
      fw.write("}");
    }

    // bypass the key
    p += strlen(key) + 3;
  }
  else {
    // bypass '@'
    fw.write("@");
    p += 1;
  }
}

void UmlItem::write() {
  if (refId != 0) {
    fw.write("<a href=\"");
    if ((parent() != 0) && (parent()->kind() == aClass)){
      fw.write("class");
      fw.write(parent()->refId);
    }
    else
      fw.write("index");
    fw.write(".html#ref");
    fw.write(refId);
    fw.write("\"><b>");
    writeq(pretty_name());
    fw.write("</b></a>");
  }
  else
    writeq(name());
}

void UmlItem::write(QCString target) {
  if (refId != 0) {
    fw.write("<a href=\"index.html#ref");
    fw.write(refId);
    fw.write("\" target = \"");
    fw.write(target);
    fw.write("\"><b>");
    writeq(pretty_name());
    fw.write("</b></a>");
  }
  else
    writeq(name());
}

void UmlItem::writeq(QCString s)
{
  const char * b = s;
  
  for (const char * p = b; *p; p += 1) {
    switch (*p) {
    case '<':
      fw.write("&lt;");
      break;
    case '>':
      fw.write("&gt;");
      break;
    case '&':
      fw.write("&amp;");
      break;
    case '':
      fw.write("&eacute;");
      break;
    case '':
      fw.write("&egrave;");
      break;
    case '':
      fw.write("&ccedil;");
      break;
    case '':
      fw.write("&agrave;");
      break;
    case '':
      fw.write("&ugrave;");
      break;
    case '':
      fw.write("&micro;");
      break;
    case '':
      fw.write("&sect;");
      break;
    case '\n':
      if ((p != b) && (p[- 1] == '\n'))
	fw.write("<br>");
      else
	fw.write(' ');
      break;
    default:
      fw.write(*p);
      break;
    }
  }
}

void UmlItem::write(const UmlTypeSpec & t, bool cpp)
{
  if (t.type != 0)
    t.type->write();
  else
    writeq((cpp) ? CppSettings::type(t.toString())
		 : JavaSettings::type(t.toString()));

}

void UmlItem::write(aVisibility v)
{
  switch (v) {
  case PublicVisibility:
    fw.write("public");
    break;
  case ProtectedVisibility:
    fw.write("protected");
    break;
  case PrivateVisibility:
    fw.write("private");
    break;
  default:
    fw.write("???");
  }
}

void UmlItem::generate_index(Vector & v, QCString k, QCString r)
{
  int n = v.size();
  
  if (n != 0) {
    sort(v);
    
    start_file(r, k + " Index", TRUE, 0);
    
    fw.write("<table>\n");
    
    for (int i = 0; i != n; i += 1) {
      UmlItem * x = v.elementAt(i);
      
      fw.write("<tr bgcolor=#f0f0f0><td>");
      x->write();
      fw.write("</td><td>");
      writeq(x->description());
      fw.write("</td></tr>\n");
    }
    fw.write("</table>\n");
    
    end_file();
  }
}

void UmlItem::sort(Vector & v)
{
  sort(v, 0, v.size() - 1);
}

void UmlItem::sort(Vector & v, int low, int high)
{
  if (low < high) {
    int lo = low;
    int hi = high + 1;
    UmlItem * e = v.elementAt(low);
    
    for (;;) {
      while ((++lo < hi) && !v.elementAt(lo)->gt(e))
	;
      while (v.elementAt(--hi)->gt(e));
	;
      
      if (lo < hi) {
	UmlItem * x = v.elementAt(lo);
	
	v.setElementAt(v.elementAt(hi), lo);
	v.setElementAt(x, hi);
      }
      else
	break;
    }
    
    UmlItem * x = v.elementAt(low);
	
    v.setElementAt(v.elementAt(hi), low);
    v.setElementAt(x, hi);
    
    sort(v, low, hi - 1);
    sort(v, hi + 1, high);
  }
}

bool UmlItem::gt(UmlItem * other) {
 QCString s1 = pretty_name();
 QCString s2 = other->pretty_name();
 int i = qstricmp((const char *) s1, (const char *) s2);
 
 if ((i == 0) && (parent() != 0) && (other->parent() != 0)) {
   s1 = parent()->pretty_name();
   s2 = other->parent()->pretty_name();
   
   return (qstricmp((const char *) s1, (const char *) s2) > 0);
 }
 else
  return (i > 0);
}

QCString UmlItem::pretty_name() {
  return name().isEmpty() ? sKind() : name();
}

Vector UmlItem::all;

FileWriter UmlItem::fw;

QCString UmlItem::directory;

unsigned int UmlItem::nrefs = 0;

QCString UmlItem::letters;

