
#include <stdio.h>
#include <qfile.h>

#include "UmlCom.h"
#include "Dialog.h"

#include "UmlPackage.h"
#include "FileOut.h"

void UmlPackage::xmi() {
  if (parent() != 0) {
    UmlItem::xmi();
    return;
  }

  // the project

  QCString path;
  QCString genviews;

  propertyValue("gxmi path", path);
  propertyValue("gxmi gen views", genviews);

  Dialog dialog(path, genviews);
  
  if (dialog.exec() != QDialog::Accepted)
    return;
    
  if (! path.isEmpty()) {
    // note : QTextStream(FILE *) bugged under windows
    QFile fp(path);
    
    if (! fp.open(IO_WriteOnly))
      UmlCom::trace("cannot open " + path);
    else {
      set_PropertyValue("gxmi path", path);
      set_PropertyValue("gxmi gen views", genviews);
      _gen_views = (genviews == "yes");
	
      {
	FileOut out(&fp);
	
	// header

#ifdef WIN32
	out << "<?xml version=\"1.0\"  encoding=\"windows-1252\"?>\n";
#else
	out << "<?xml version=\"1.0\"  encoding=\"ISO-8859\"?>\n";
#endif
	out << "\
<XMI xmi.version=\"1.2\" xmlns:UML=\"org.omg.xmi.namespace.UML\">\n\
<XMI.header>\n\
	<XMI.documentation>\n\
		<XMI.exporter>Bouml</XMI.exporter>\n\
		<XMI.exporterVersion>1.0</XMI.exporterVersion>\n\
	</XMI.documentation>\n\
	<XMI.metamodel XMI.name=\"UML\" XMI.version=\"1.4\"/>\n\
</XMI.header>\n\
<XMI.content>\n\
	<UML:Model name=\"" << name() << '"';
	out.id(this);
	out << ">\n\
		<UML:Namespace.ownedElement>\n";
	out.indent(+3);
	
	_written = TRUE;
	
	// write children
	
	const QVector<UmlItem> ch = children();
	unsigned n = ch.size();
  
	for (unsigned i = 0; i != n; i += 1)
	  ch[i]->write_if_needed(out);
	
	// stereotypes, data types etc ...
#if 0
	out.define_stereotypes();
#endif
	out.define_datatypes();
	
#if 0
	out.indent();
	out << "<UML:TagDefinition xmi.id=\"BOUML_tag_description\" name=\"documentation\" isSpecification=\"false\" tagType=\"String\"/>\n";
#endif

	// footer
	
	out << "\
		</UML:Namespace.ownedElement>\n\
	</UML:Model>\n\
</XMI.content>\n\
</XMI>\n";
      }
      
      UmlCom::trace("xmi generation done in <i>" + path + "</i><br>");
      
      fp.flush();
      fp.close();
    }
  }
}

bool UmlPackage::write_if_needed(FileOut & out) {
  // always write package header because may be the target of a relation
  parent()->write(out);
  out.indent(); 
  out << "<UML:Package"; 
  out.id(this); 
  out << " name =\"" << name() << '"'; 
  out << ">\n"; 
  out.indent(+1); 
  write_stereotype(out); 
  write_description(out);   

  const QVector<UmlItem> ch = children();
  unsigned n = ch.size();
  bool used = FALSE;
  
  for (unsigned i = 0; i != n; i += 1)
    used |= ch[i]->write_if_needed(out);
    
  if (used) {
    out.indent(-1);
out.indent(); out << "<!-- for pack " << name() << "-->\n";
    out.indent();
    out << "</UML:Namespace.ownedElement>\n";
  }

  out.indent(-1); 
  out.indent(); 
  out << "</UML:Package>\n"; 

  return TRUE;
}

void UmlPackage::write(FileOut & out) {
  if (!_written) {
    _written = TRUE;
    out.indent();
    out << "<UML:Namespace.ownedElement>\n";
    out.indent(+1);
  }

}

