// *************************************************************************
//
// Copyright (C) 2004,2005 Bruno PAGES  All rights reserved.
//
// This file is part of the BOUML Uml Toolkit.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
// e-mail : bouml@free.fr
// home   : http://bouml.free.fr
//
// *************************************************************************

#include <qpopupmenu.h> 
#include <qmessagebox.h>
#include <qfile.h> 
#include <qcursor.h>
#include <qapplication.h>
#include <qfiledialog.h>

#include "BrowserPackage.h"
#include "PackageData.h"
#include "BrowserClassView.h"
#include "BrowserUseCaseView.h"
#include "BrowserClassDiagram.h"
#include "BrowserClass.h"
#include "BrowserUseCase.h"
#include "BrowserView.h"
#include "BrowserAttribute.h"
#include "BrowserOperation.h"
#include "BrowserComponent.h"
#include "BrowserArtifact.h"
#include "BrowserDeploymentNode.h"
#include "BrowserComponentView.h"
#include "BrowserDeploymentView.h"
#include "BrowserSimpleRelation.h"
#include "BrowserState.h"
#include "BrowserPseudoState.h"
#include "BrowserStateAction.h"
#include "SimpleRelationData.h"
#include "UmlPixmap.h"
#include "UmlDrag.h"
#include "SettingsDialog.h"
#include "StereotypesDialog.h"
#include "GenerationSettings.h"
#include "myio.h"
#include "ToolCom.h"
#include "Tool.h"
#include "MenuTitle.h"
#include "UmlWindow.h"
#include "ReferenceDialog.h"
#include "UmlGlobal.h"
#include "mu.h"
#include "SaveProgress.h"

IdDict<BrowserPackage> BrowserPackage::all;
QList<BrowserPackage> BrowserPackage::removed;

QStringList BrowserPackage::its_default_stereotypes;
QStringList BrowserPackage::relation_default_stereotypes;

BrowserPackage::BrowserPackage(QString s, BrowserView * parent, int id)
    : BrowserNode(s, parent), Labeled<BrowserPackage>(all, id),
      def(new PackageData), associated_diagram(0), is_imported(FALSE) {
  // creates the project package
  is_modified = (id == 0);
  def->set_browser_node(this);
  
  classdiagram_settings.draw_all_relations = UmlYes;
  classdiagram_settings.hide_attributes = UmlNo;
  classdiagram_settings.hide_operations = UmlNo;
  classdiagram_settings.show_full_members_definition = UmlNo;
  classdiagram_settings.package_name_in_tab = UmlNo;
  classdiagram_settings.class_drawing_mode = Natural;
  classdiagram_settings.drawing_language = UmlView;
  classdiagram_settings.show_context_mode = noContext;
  classdiagram_settings.auto_label_position = UmlYes;
    
  class_settings.attribute_visibility = UmlProtected;
  class_settings.relation_visibility = UmlProtected;
  class_settings.operation_visibility = UmlPublic;
    
  usecasediagram_settings.package_name_in_tab = UmlNo;
  usecasediagram_settings.show_context_mode = noContext;
  usecasediagram_settings.auto_label_position = UmlYes;
  
  sequencediagram_settings.show_full_operations_definition = UmlNo;
  sequencediagram_settings.write_horizontally = UmlYes;
  sequencediagram_settings.drawing_language = UmlView;
  
  collaborationdiagram_settings.show_full_operations_definition = UmlNo;
  collaborationdiagram_settings.show_hierarchical_rank = UmlNo;
  collaborationdiagram_settings.write_horizontally = UmlYes;
  collaborationdiagram_settings.drawing_language = UmlView;
  
  componentdiagram_settings.package_name_in_tab = UmlNo;
  componentdiagram_settings.show_context_mode = noContext;
  componentdiagram_settings.auto_label_position = UmlYes;
  componentdiagram_settings.componentdrawingsettings.draw_component_as_icon = UmlNo;
  componentdiagram_settings.componentdrawingsettings.show_component_req_prov = UmlNo;
  componentdiagram_settings.componentdrawingsettings.show_component_rea = UmlNo;
  
  deploymentdiagram_settings.package_name_in_tab = UmlNo;
  deploymentdiagram_settings.show_context_mode = noContext;
  deploymentdiagram_settings.write_horizontally = UmlYes;
  deploymentdiagram_settings.auto_label_position = UmlYes;
  deploymentdiagram_settings.componentdrawingsettings.draw_component_as_icon = UmlNo;
  deploymentdiagram_settings.componentdrawingsettings.show_component_req_prov = UmlNo;
  deploymentdiagram_settings.componentdrawingsettings.show_component_rea = UmlNo;
  
  statediagram_settings.package_name_in_tab = UmlNo;
  statediagram_settings.show_context_mode = noContext;
  statediagram_settings.auto_label_position = UmlYes;
  statediagram_settings.write_trans_label_horizontally = UmlYes;
  statediagram_settings.statedrawingsettings.show_activities = UmlYes;
  statediagram_settings.statedrawingsettings.region_horizontally = UmlYes;
  statediagram_settings.statedrawingsettings.drawing_language = UmlView;
  
  class_color = UmlYellow;
  package_color = UmlWhite;
  fragment_color = UmlWhite;
  note_color = UmlBlue;
  usecase_color = UmlYellow;
  component_color = UmlGreen;
  artifact_color = UmlGreen;
  deploymentnode_color = UmlGray;
  duration_color = UmlWhite;
  continuation_color = UmlGray;
  state_color = UmlYellow;
  stateaction_color = UmlWhite;
}

BrowserPackage::BrowserPackage(QString s, BrowserNode * parent, int id)
    : BrowserNode(s, parent), Labeled<BrowserPackage>(all, id),
      def(new PackageData), associated_diagram(0), is_imported(FALSE) {
  make();
  is_modified = (id == 0);
}

BrowserPackage::BrowserPackage(int id)
    : Labeled<BrowserPackage>(all, id),
      def(new PackageData), associated_diagram(0), is_imported(FALSE) {
  make();
  is_modified = (id == 0);
}

BrowserPackage::~BrowserPackage() {
  if (this == BrowserView::get_project()) {
    // to accelerate removals
    removed.clear();
    clear(FALSE);
  }
  else {
    removed.remove(this);
    all.remove(get_ident());
  }
  delete def;
}

void BrowserPackage::make() {
  def->set_browser_node(this);
  
  class_color = UmlDefaultColor;
  note_color = UmlDefaultColor;
  package_color = UmlDefaultColor;
  fragment_color = UmlDefaultColor;
  duration_color = UmlDefaultColor;
  continuation_color = UmlDefaultColor;
  usecase_color = UmlDefaultColor;
  component_color = UmlDefaultColor;
  artifact_color = UmlDefaultColor;
  deploymentnode_color = UmlDefaultColor;
  state_color = UmlDefaultColor;
  stateaction_color = UmlDefaultColor;
}

void BrowserPackage::clear(bool old)
{
  if (! old)
    set_user_id(-1);
  
  all.clear(old);
  BrowserClassView::clear(old);
  BrowserUseCaseView::clear(old);
  BrowserComponentView::clear(old);
  BrowserDeploymentView::clear(old);
  BrowserDiagram::clear(old);
}

void BrowserPackage::update_idmax_for_root()
{
  all.update_idmax_for_root();
  BrowserClassView::update_idmax_for_root();
  BrowserUseCaseView::update_idmax_for_root();
  BrowserComponentView::update_idmax_for_root();
  BrowserDeploymentView::update_idmax_for_root();
  BrowserDiagram::update_idmax_for_root();
}
    
const QPixmap* BrowserPackage::pixmap(int) const {
  return (this != BrowserView::get_project())
    ? ((deletedp()) ? DeletedPackageIcon : PackageIcon)
    : 0;
}

QString BrowserPackage::full_name(bool rev, bool) const {
  if (this == BrowserView::get_project())
    return QString::null;
  
  QString p = ((BrowserNode *) parent())->full_name(FALSE, FALSE);

  if (p.isEmpty()) 
    return QString((const char *) name);
  else if (rev)
    return name + "   [" + p + "]";
  else
    return p + "::" + name;
}

// just check if the inheritance already exist
const char * BrowserPackage::check_inherit(BrowserNode * new_parent) {
  QListViewItem * child;
    
  for (child = firstChild(); child != 0; child = child->nextSibling()) {
    BrowserNode * ch = ((BrowserNode *) child);
    
    if ((ch->get_type() == UmlInherit) && 
	((((SimpleRelationData *) ch->get_data())->get_end_node())
	 == new_parent))
      return "already exist";
  }
  
  return (new_parent != this) ? 0 : "circular inheritance";
}

void BrowserPackage::menu() {
  QPopupMenu m(0);
  QPopupMenu genm(0);
  QPopupMenu revm(0);
  QPopupMenu toolm(0);
  QPopupMenu importm(0);
  
  m.insertItem(new MenuTitle(name, m.font()), -1);
  m.insertSeparator();
  if (!deletedp()) {
    if (!is_read_only && (edition_number == 0)) {
      m.setWhatsThis(m.insertItem("new package", 0),
		     "to add a new sub <em>package</em>");
      m.setWhatsThis(m.insertItem("new use case view", 1),
		     "to add a new <em>use case view</em>");
      m.setWhatsThis(m.insertItem("new class view", 2),
		     "to add a new <em>class view</em>");
      m.setWhatsThis(m.insertItem("new component view", 3),
		     "to add a new <em>class view</em>");
      m.setWhatsThis(m.insertItem("new deployment view", 4),
		     "to add a new <em>deployment view</em>");
      m.setWhatsThis(m.insertItem("import project", 14),
		     "to import the contents of a <em>project</em> under \
the current <em>package</em>");
      m.insertSeparator();
      m.insertSeparator();
    }
    if (!is_edited) {
      m.setWhatsThis(m.insertItem("edit", 5),
		     "to edit the package");
      if (!is_read_only) {
	m.insertSeparator();
	if (this == BrowserView::get_project())
	  m.setWhatsThis(m.insertItem("edit default stereotypes", 6),
			 "to set the default stereotypes list");
	m.setWhatsThis(m.insertItem("edit class settings", 7),
		       "to set the sub classes's settings");
	if (this == BrowserView::get_project())
	  m.setWhatsThis(m.insertItem("edit generation settings", 8),
			 "to set how an Uml type is compiled in C++ etc..., \
to set the default parameter passing, to set the default code \
produced for an attribute etc..., and to set the root directories");
	m.setWhatsThis(m.insertItem("edit drawing settings", 9),
		       "to set how the sub <em>diagrams</em>'s items must be drawed");
	if (this == BrowserView::get_project()) {
	  m.insertItem("import", &importm);
	  m.setWhatsThis(importm.insertItem("generation settings", 28),
			 "to import the generation settings from an other project, \
note that the root directories are not imported");
	  m.setWhatsThis(importm.insertItem("default stereotypes", 29),
			 "to import the default stereotypes from an other project");
	}
	if ((edition_number == 0) && (this != BrowserView::get_project())) {
	  m.insertSeparator();
	  m.setWhatsThis(m.insertItem("delete", 10),
			 "to delete the <em>package</em> and its sub items. \
Note that you can undelete it after");
	}
      }
    }
    m.insertSeparator();
    m.setWhatsThis(m.insertItem("referenced by", 13),
		   "to know who reference the <i>package</i> \
through a relation");
    mark_menu(m, "package", 90);
    m.insertSeparator();
    m.insertItem("generate", &genm);
    genm.insertItem("C++", 20);
    genm.insertItem("Java", 21);
    genm.insertItem("Idl", 22);
    
    if (edition_number == 0) {
      if (!is_read_only) {
	m.insertItem("reverse", &revm);
	
	revm.insertItem("C++", 24);
	revm.insertItem("Java", 25);
	revm.insertItem("Java Catalog", 26);
      }

      if (Tool::menu_insert(&toolm, 
			    (this == BrowserView::get_project()) ? UmlProject : UmlPackage,
			    100)) {
	m.insertSeparator();
	m.insertItem("tool", &toolm);
      }
    }
  }
  else if (!is_read_only && (edition_number == 0)) {
    m.setWhatsThis(m.insertItem("undelete", 11),
		   "undelete the <em>package</em>. Do not undelete its sub items");
    m.setWhatsThis(m.insertItem("undelete recursively", 12),
		   "undelete the <em>package</em> and its sub items");
  }
  
  static int phase_renumerotation = -1;
  
  if (user_id() == 0) {
    m.insertSeparator();
    m.insertSeparator();
    m.insertSeparator();
    m.insertSeparator();
    m.insertSeparator();
    
    switch (phase_renumerotation) {
    case -1:
      if (this == BrowserView::get_project())
	m.insertItem("PREPARE la renumerotation", 27);
      break;
    case 0:
      m.insertItem("Renumerotation BASE", 27);
      break;
    case 1:
      m.insertItem("Renumerotation USER", 27);
      break;
    }
  }
  
  int rank = m.exec(QCursor::pos());
  
  switch (rank) {
  case 0:
    add_package();
    break;
  case 1:
    add_use_case_view();
    break;
  case 2:
    add_class_view();
    break;
  case 3:
    add_component_view();
    break;
  case 4:
    add_deployment_view();
    break;
  case 14:
    import_project();
    break;
  case 5:
    def->edit();
    return;
  case 6:
    {
      StereotypesDialog d(QCursor::pos());
      
      if (! d.exec())
	return;
      break;
    }
  case 7:
    if (!class_settings.edit((this == BrowserView::get_project())
			     ? UmlProject : UmlPackage))
      return;
    break;
  case 8:
    if (!GenerationSettings::edit())
      return;
    break;
  case 9:
    {
      QArray<StateSpec> st;
      QArray<ColorSpec> co(12);
      
      classdiagram_settings.complete(st, UmlClassView);
      sequencediagram_settings.complete(st, FALSE);
      collaborationdiagram_settings.complete(st, FALSE);
      
      usecasediagram_settings.complete(st, FALSE);
      componentdiagram_settings.complete(st, FALSE);
      deploymentdiagram_settings.complete(st, FALSE);
      statediagram_settings.complete(st, FALSE);
      
      co[0].set("class color", &class_color);
      co[1].set("note color", &note_color);
      co[2].set("package color", &package_color);
      co[3].set("fragment color", &fragment_color);
      co[4].set("use case color", &usecase_color);
      co[5].set("duration color", &duration_color);
      co[6].set("continuation color", &continuation_color);
      co[7].set("component color", &component_color);
      co[8].set("artifact color", &artifact_color);
      co[9].set("node color", &deploymentnode_color);
      co[10].set("state color", &state_color);
      co[11].set("state action color", &stateaction_color);

      SettingsDialog dialog(&st, &co, this == BrowserView::get_project(), FALSE);
      
      if (dialog.exec() != QDialog::Accepted)
	return;
      DrawingSettings::modified();
    }
    break;
  case 10:
    delete_it();
    ((BrowserNode *) parent())->package_modified();
    break;
  case 11:
    undelete(FALSE);
    break;
  case 12:
    undelete(TRUE);
    break;
  case 13:
    ReferenceDialog::show(this);
    return;
  case 20:
    ToolCom::run("cpp_generator", this);
    return;
  case 21:
    ToolCom::run("java_generator", this);
    return;
  case 22:
    ToolCom::run("idl_generator", this);
    return;
  case 24:
    ToolCom::run("cpp_reverse", this);
    return;
  case 25:
    ToolCom::run("java_reverse", this);
    return;
  case 26:
    ToolCom::run("java_catalog", this);
    return;
  case 27:
    renumber(phase_renumerotation++);
    if (phase_renumerotation == 2) {
      renumber(2);
#if 0
      ???
      all.insert(BrowserView::get_project()->get_ident(), 
		 BrowserView::get_project());
#endif
      save_all(FALSE);
      QApplication::exit();
    }
    return;
  case 28:
    if (!GenerationSettings::import())
      return;
  case 29:
    if (!import_stereotypes())
      return;
  default:
    if (rank >= 100)
      ToolCom::run(Tool::command(rank - 100), this);
    else
      mark_management(rank - 90);
    return;
  }
  
  package_modified();
}

BrowserNodeList BrowserPackage::instances() {
  IdIterator<BrowserPackage> it(all);
  BrowserNodeList result;
  
  while (it.current() != 0) {
    if (!it.current()->deletedp())
      result.append(it.current());
    ++it;
  }

  result.sort();
  
  return result;
}

BrowserPackage * BrowserPackage::get_package()
{
  if (BrowserView::get_project()->firstChild() == 0)
    // only the prj
    return 0;
  
  BrowserNode * old;
  QString dummy;
  BrowserView::get_project()->enter_child_name(dummy, "choose existing package : ",
					       UmlPackage, instances(), &old,
					       FALSE, TRUE, TRUE);

  return ((BrowserPackage *) old);
}

void BrowserPackage::add_package() {
  (new BrowserPackage(child_random_name("Package"), this))
    ->select_in_browser();
}

void BrowserPackage::add_use_case_view() {
  (new BrowserUseCaseView(child_random_name("Use Case View"), this))
    ->select_in_browser();
}

void BrowserPackage::add_class_view() {
  (new BrowserClassView(child_random_name("Class view"), this))
    ->select_in_browser();
}

void BrowserPackage::add_component_view() {
  (new BrowserComponentView(child_random_name("Component view"), this))
    ->select_in_browser();
}

void BrowserPackage::add_deployment_view() {
  (new BrowserDeploymentView(child_random_name("Deployment view"), this))
    ->select_in_browser();
}

void BrowserPackage::import_project() {
  QString fn = QFileDialog::getOpenFileName(QString::null, "*.prj");
  
  if (!fn.isEmpty()) {
    QFileInfo fi(fn);
  
    if (!fi.exists())
      return;

    if (wrong_child_name(fi.baseName(), UmlPackage, TRUE, FALSE)) {
      QMessageBox::critical(0, "Error", "illegal name or already used");
      return;
    }
  
    QApplication::setOverrideCursor(Qt::waitCursor);
    QDir di(fi.dirPath());
    
    BrowserPackage * p = new BrowserPackage(fi.baseName(), this);
    bool err = FALSE;
    
    set_in_import(TRUE);
    BrowserView::set_imported_project(di, p);
    BrowserNode::pre_load();
    try {
      p->load(TRUE);
    }
    catch (int) {
      err = TRUE;
    }
    BrowserNode::post_load();
    BrowserDiagram::import();
    BrowserView::set_imported_project(di, 0);
    set_in_import(FALSE);
    clear(TRUE);
    QApplication::restoreOverrideCursor();
    
    if (err) {
      p->delete_it();
      delete p;
    }
  }
}


UmlCode BrowserPackage::get_type() const {
  return UmlPackage;
}

BasicData * BrowserPackage::get_data() const {
  return def;
}

void BrowserPackage::get_classdiagramsettings(ClassDiagramSettings & r) const {
  if (! classdiagram_settings.complete(r))
    ((BrowserNode *) parent())->get_classdiagramsettings(r);
}

void BrowserPackage::get_sequencediagramsettings(SequenceDiagramSettings & r) const {
  if (! sequencediagram_settings.complete(r))
    ((BrowserNode *) parent())->get_sequencediagramsettings(r);
}

void BrowserPackage::get_collaborationdiagramsettings(CollaborationDiagramSettings & r) const {
  if (! collaborationdiagram_settings.complete(r))
    ((BrowserNode *) parent())->get_collaborationdiagramsettings(r);
}

void BrowserPackage::get_usecasediagramsettings(UseCaseDiagramSettings & r) const {
  if (! usecasediagram_settings.complete(r))
    ((BrowserNode *) parent())->get_usecasediagramsettings(r);
}

void BrowserPackage::get_componentdiagramsettings(ComponentDiagramSettings & r) const {
  if (! componentdiagram_settings.complete(r))
    ((BrowserNode *) parent())->get_componentdiagramsettings(r);
}

void BrowserPackage::get_deploymentdiagramsettings(DeploymentDiagramSettings & r) const {
  if (! deploymentdiagram_settings.complete(r))
    ((BrowserNode *) parent())->get_deploymentdiagramsettings(r);
}

void BrowserPackage::get_statediagramsettings(StateDiagramSettings & r) const {
  if (! statediagram_settings.complete(r))
    ((BrowserNode *) parent())->get_statediagramsettings(r);
}

UmlColor BrowserPackage::get_color(UmlCode who) const {
  UmlColor c;
  
  switch (who) {
  case UmlClass:
    c = class_color;
    break;
  case UmlNote:
    c = note_color;
    break;
  case UmlFragment:
    c = fragment_color;
    break;
  case UmlPackage:
    c = package_color;
    break;
  case UmlUseCase:
    c = usecase_color;
    break;
  case UmlComponent:
    c = component_color;
    break;
  case UmlArtifact:
    c = artifact_color;
    break;
  case UmlDeploymentNode:
    c = deploymentnode_color;
    break;
  case UmlContinuation:
    c = continuation_color;
    break;
  case UmlActivityDuration:
    c = duration_color;
    break;
  case UmlState:
    c = state_color;
    break;
  default:	// UmlStateAction
    c = stateaction_color;
  }
  
  return (c != UmlDefaultColor)
    ? c
    : ((BrowserNode *) parent())->get_color(who);
}

bool BrowserPackage::get_classinstwritehorizontally(bool seq) const {
  Uml3States h = (seq)
    ? sequencediagram_settings.write_horizontally
    : collaborationdiagram_settings.write_horizontally;
  
  return (h == UmlDefaultState)
    ? ((BrowserNode *) parent())->get_classinstwritehorizontally(seq)
    : (h == UmlYes);
}

void BrowserPackage::get_componentdrawingsettings(bool depl, ComponentDrawingSettings & result) const {
  if ((depl) 
      ? !deploymentdiagram_settings.componentdrawingsettings.complete(result)
      : !componentdiagram_settings.componentdrawingsettings.complete(result))
    ((BrowserNode *) parent())->get_componentdrawingsettings(depl, result);
}

void BrowserPackage::get_statedrawingsettings(StateDrawingSettings & result) const {
  if (!statediagram_settings.statedrawingsettings.complete(result))
    ((BrowserNode *) parent())->get_statedrawingsettings(result);
}

UmlVisibility BrowserPackage::get_visibility(UmlCode who) const {
  UmlVisibility v;
  
  switch (who) {
  case UmlAttribute:
    v = class_settings.attribute_visibility;
    break;
  case UmlRelations:
    v = class_settings.relation_visibility;
    break;
  default:	// UmlOperation
    v = class_settings.operation_visibility;
    break;
  }
  
  return (v != UmlDefaultVisibility)
    ? v
    : ((BrowserNode *) parent())->get_visibility(who);
}

bool BrowserPackage::get_auto_label_position(UmlCode who) const {
  Uml3States v;
  
  switch (who) {
  case UmlClassDiagram:
    v = classdiagram_settings.auto_label_position;
    break;
  case UmlUseCaseDiagram:
    v = usecasediagram_settings.auto_label_position;
    break;
  case UmlStateDiagram:
    v = statediagram_settings.auto_label_position;
    break;
  case UmlComponentDiagram:
    v = componentdiagram_settings.auto_label_position;
    break;
  case UmlDeploymentDiagram:
    v = deploymentdiagram_settings.auto_label_position;
    break;
  default:
    // error
    return FALSE;
  }
  
  switch (v) {
  case UmlYes:
    return TRUE;
  case UmlNo:
    return FALSE;
  default:
    return ((BrowserNode *) parent())->get_auto_label_position(who);
  }
}

bool BrowserPackage::get_write_transition_label_horizontally(UmlCode who) const {
  Uml3States v;
  
  switch (who) {
  case UmlStateDiagram:
    v = statediagram_settings.auto_label_position;
    break;
  default:
    // error
    return FALSE;
  }
  
  switch (v) {
  case UmlYes:
    return TRUE;
  case UmlNo:
    return FALSE;
  default:
    return ((BrowserNode *) parent())->get_write_transition_label_horizontally(who);
  }
}

DrawingLanguage BrowserPackage::get_language(UmlCode who) const {
  DrawingLanguage v;
  
  switch (who) {
  case UmlStateDiagram:
    v = statediagram_settings.statedrawingsettings.drawing_language;
    break;
  default:
    // error
    return UmlView;
  }
  
  return (v != DefaultDrawingLanguage)
    ? v
    : ((BrowserNode *) parent())->get_language(who);
}

void BrowserPackage::open(bool force_edit) {
  if (!force_edit &&
      (associated_diagram != 0) &&
      !associated_diagram->deletedp())
    associated_diagram->open(FALSE);
  else if (!is_edited)
    def->edit();
}

void BrowserPackage::on_close() {
}

BrowserNode * BrowserPackage::get_associated() const {
  return associated_diagram;
}

void BrowserPackage::set_associated_diagram(BrowserNode * dg,
					    bool on_read) {
  if (associated_diagram != dg) {
    if (associated_diagram != 0)
      QObject::disconnect(associated_diagram->get_data(), SIGNAL(deleted()),
			  def, SLOT(on_delete()));
    if ((associated_diagram = dg) != 0)
      QObject::connect(associated_diagram->get_data(), SIGNAL(deleted()),
		       def, SLOT(on_delete()));
    
    if (! on_read)
      package_modified();
  }
}

void BrowserPackage::on_delete() {
  if (associated_diagram && associated_diagram->deletedp())
    associated_diagram = 0;
}

void BrowserPackage::write_id(ToolCom * com) {
  // to manage project case as any package
  com->write_id(this, UmlPackage - UmlRelations, name);
}

bool BrowserPackage::tool_cmd(ToolCom * com, const char * args) {
  switch ((unsigned char) args[-1]) {
  case supportFileCmd:
    {
      QString fn;
      
      if (this == BrowserView::get_project())
	fn.sprintf("%s.prj", (const char *) name);
      else 
	fn.setNum(get_ident());

      QDir d = BrowserView::get_dir();
      
      com->write_string(d.absFilePath(fn));
    }
    return TRUE;
  case createCmd:
    {
      bool ok = TRUE;
      
      if (is_read_only && !root_permission())
	ok = FALSE;
      else {
	switch ((UmlCode) ((int) *args++ + UmlRelations)) {
	case UmlPackage:
	  if (wrong_child_name(args, UmlPackage, TRUE, FALSE))
	    ok = FALSE;
	  else
	    (new BrowserPackage(args, this))->write_id(com);
	  break;
	case UmlSimpleRelations:
	  {
	    UmlCode c;
	    
	    if (!com->get_relation_kind(c, args))
	      ok = FALSE;
	    else {
	      BrowserNode * end = (BrowserNode *) com->get_id(args);
	      
	      if (end->get_type() == UmlPackage) {
		switch (c) {
		case UmlDependency:
		  add_relation(UmlDependOn, end)->get_browser_node()->write_id(com);
		case UmlGeneralisation:
		  if (check_inherit(end) == 0)
		    add_relation(UmlInherit, end)->get_browser_node()->write_id(com);
		  else
		    ok = FALSE;
		  break;
		default:
		  ok = FALSE;
		}
	      }
	      else
		ok = FALSE;
	    }
	  }
	  break;
	case UmlUseCaseView:
	  if (wrong_child_name(args, UmlUseCaseView, TRUE, FALSE))
	    ok = FALSE;
	  else
	    (new BrowserUseCaseView(args, this))->write_id(com);
	  break;
	case UmlClassView:
	  if (wrong_child_name(args, UmlClassView, TRUE, FALSE))
	    ok = FALSE;
	  else
	    (new BrowserClassView(args, this))->write_id(com);
	  break;
	case UmlComponentView:
	  if (wrong_child_name(args, UmlComponentView, TRUE, FALSE))
	    ok = FALSE;
	  else
	    (new BrowserComponentView(args, this))->write_id(com);
	  break;
	case UmlDeploymentView:
	  if (wrong_child_name(args, UmlDeploymentView, TRUE, FALSE))
	    ok = FALSE;
	  else
	    (new BrowserDeploymentView(args, this))->write_id(com);
	  break;
	default:
	  ok = FALSE;
	}
      }
        
      if (! ok)
	com->write_id(0);
      else
	package_modified();
      
      return TRUE;
    }
    break;
  case setAssocDiagramCmd:
    if (is_read_only && !root_permission())
      com->write_bool(FALSE);
    else {
      set_associated_diagram((BrowserClassDiagram *) com->get_id(args));
      com->write_bool(TRUE);
    }
    return TRUE;
  default:
    return (def->tool_cmd(com, args, this, comment) ||
	    BrowserNode::tool_cmd(com, args));
  }
}

bool BrowserPackage::tool_global_cmd(ToolCom * com, const char * args)
{
  const MyStr & (PackageData::* pf)() const;
  
  switch ((unsigned char) args[-1]) {
  case findNamespaceCmd:
    pf = &PackageData::get_cpp_namespace;
    break;
  case findPackageCmd:
    pf = &PackageData::get_java_package;
    break;
  case findModuleCmd:
    pf = &PackageData::get_idl_module;
    break;
  case getProjectCmd:
    BrowserView::get_project()->write_id(com);
    return TRUE;
  case isProjectModifiedCmd:
    com->write_char(must_be_saved() != FALSE);
    return TRUE;
  case saveProjectCmd:
    UmlWindow::save_it();
    return TRUE;
  default:
    return FALSE;
  }
  
  // search for a package/namespace/module
  BrowserPackage * p = (BrowserPackage *) com->get_id(args);
  
  if (p == 0)
    // old version
    p = BrowserView::get_project();
  
  if ((p = p->find_it(args, pf)) != 0)
    p->write_id(com);
  else
    com->write_id(0);
  
  return TRUE;
}

BrowserPackage *
  BrowserPackage::find_it(const char * s,
			  const MyStr & (PackageData::* pf)() const) {
  if (!deletedp()) {
    if ((def->*pf)() == s) {
      // find !
      return this;
    }
    else {
      for (QListViewItem * child = firstChild();
	   child != 0;
	   child = child->nextSibling()) {
	if (((BrowserNode *) child)->get_type() == UmlPackage) {
	  BrowserPackage * p = 
	    ((BrowserPackage *) child)->find_it(s, pf);
	  
	  if (p != 0)
	    return p;
	}
      }
    }
  }
  
  return 0;
}


void BrowserPackage::DragMoveEvent(QDragMoveEvent * e) {
  if (UmlDrag::canDecode(e, UmlPackage) ||
      UmlDrag::canDecode(e, UmlUseCaseView) ||
      UmlDrag::canDecode(e, UmlClassView) ||
      UmlDrag::canDecode(e, UmlComponentView) ||
      UmlDrag::canDecode(e, UmlDeploymentView) ||
      UmlDrag::canDecode(e, BrowserSimpleRelation::drag_key(this)))
    e->accept();
  else if (parent() != 0)
    ((BrowserNode *) parent())->DragMoveInsideEvent(e);
  else
    e->ignore();
}

void BrowserPackage::DragMoveInsideEvent(QDragMoveEvent * e) {
  if (UmlDrag::canDecode(e, UmlPackage) ||
      UmlDrag::canDecode(e, UmlUseCaseView) ||
      UmlDrag::canDecode(e, UmlClassView) ||
      UmlDrag::canDecode(e, UmlComponentView) ||
      UmlDrag::canDecode(e, UmlDeploymentView) ||
      UmlDrag::canDecode(e, BrowserSimpleRelation::drag_key(this)))
    e->accept();
  else
    e->ignore();
}

bool BrowserPackage::may_contains_them(const QList<BrowserNode> & l,
				       bool & duplicable) const {
  QListIterator<BrowserNode> it(l);
  
  for (; it.current(); ++it) {
    switch (it.current()->get_type()) {
    case UmlPackage:
    case UmlUseCaseView:
    case UmlClassView:
    case UmlComponentView:
    case UmlDeploymentView:
      break;
    default:
      if (!IsaSimpleRelation(it.current()->get_type()) ||
	  (((const BrowserNode *) it.current()->parent()) != this))
	return FALSE;
    }
    
    duplicable = FALSE;
    if (! may_contains(it.current(), FALSE))
      return FALSE;
  }
  
  return TRUE;
}

void BrowserPackage::DropEvent(QDropEvent * e) {
  DropAfterEvent(e, 0);
}

void BrowserPackage::DropAfterEvent(QDropEvent * e, BrowserNode * after) {
  BrowserNode * bn;
  
  if ((((bn = UmlDrag::decode(e, UmlPackage)) != 0) ||
       ((bn = UmlDrag::decode(e, UmlUseCaseView)) != 0) ||
       ((bn = UmlDrag::decode(e, UmlClassView)) != 0) ||
       ((bn = UmlDrag::decode(e, UmlComponentView)) != 0) ||
       ((bn = UmlDrag::decode(e, UmlDeploymentView)) != 0) ||
       ((bn = UmlDrag::decode(e, BrowserSimpleRelation::drag_key(this))) != 0)) &&
      (bn != after) && (bn != this)) {
    if (may_contains(bn, bn->get_type() == UmlPackage)) {
      BrowserNode * x = this;
      
      if ((after == 0) &&
	  (bn->get_type() == UmlPackage) &&
	  (parent() != 0) &&
	  ((BrowserNode *) parent())->may_contains(bn, TRUE)) {
	// have choice
	QPopupMenu m(0);
  
	m.insertItem(new MenuTitle(bn->get_name() + QString(" moving"),
				   m.font()), -1);
	m.insertSeparator();
	m.insertItem("in " + QString(get_name()), 1);
	m.insertItem("after " + QString(get_name()), 2);
	
	switch (m.exec(QCursor::pos())) {
	case 1:
	  break;
	case 2:
	  after = this;
	  x = (BrowserNode *) parent();
	  break;
	default:
	  e->ignore();
	  return;
	}
      }
      BrowserNode * old_parent = (BrowserNode *) bn->parent();
      
      if (after)
	bn->moveItem(after);
      else {
	old_parent->takeItem(bn);
	x->insertItem(bn);
      }
      x->package_modified();
      if (old_parent != x)
	old_parent->package_modified();
    }
    else if ((parent() != 0) && (after == 0))
      ((BrowserNode *) parent())->DropAfterEvent(e, this);
    else {
      QMessageBox::critical(0, "Error", "Forbiden");
      e->ignore();
    }
  }
  else if ((bn == 0) && (after == 0))
    ((BrowserNode *) parent())->DropAfterEvent(e, this);
  else
    e->ignore();
}

// stereotypes

const QStringList & BrowserPackage::default_stereotypes()
{
  return its_default_stereotypes;
}

const QStringList & BrowserPackage::default_stereotypes(UmlCode) {
  return relation_default_stereotypes;
}

void BrowserPackage::init()
{
  its_default_stereotypes.clear();
  its_default_stereotypes.append("facade");
  its_default_stereotypes.append("framework");
  its_default_stereotypes.append("model library");
  its_default_stereotypes.append("stub");
  its_default_stereotypes.append("toplevel");
  
  relation_default_stereotypes.clear();
  relation_default_stereotypes.append("access");
  relation_default_stereotypes.append("import");
}

void BrowserPackage::save_stereotypes()
{
  QByteArray newdef;
  QTextOStream st(newdef);
	
  st.setEncoding(QTextStream::Latin1);
  
  nl_indent(st);
  st << "package_stereotypes ";
  save_string_list(its_default_stereotypes, st);
  nl_indent(st);
  st << "  " << stringify(UmlDependency);
  save_string_list(relation_default_stereotypes, st);
  nl_indent(st);
  st << "end";
  nl_indent(st);
  
  BrowserClass::save_stereotypes(st);
  BrowserUseCase::save_stereotypes(st);
  BrowserArtifact::save_stereotypes(st);
  BrowserAttribute::save_stereotypes(st);
  BrowserOperation::save_stereotypes(st);
  BrowserState::save_stereotypes(st);
  BrowserPseudoState::save_stereotypes(st);
  BrowserStateAction::save_stereotypes(st);
  BrowserComponent::save_stereotypes(st);
  BrowserDeploymentNode::save_stereotypes(st);
  BrowserClassView::save_stereotypes(st);
  BrowserUseCaseView::save_stereotypes(st);
  BrowserComponentView::save_stereotypes(st);
  BrowserDeploymentView::save_stereotypes(st);
  BrowserDiagram::save_stereotypes(st);
  st << "\nend\n";
  
  st << '\000';
  save_if_needed("stereotypes", newdef);
}

void BrowserPackage::read_stereotypes(char * & st, char * & k)
{
  if (!strcmp(k, "package_stereotypes")) {
    read_string_list(its_default_stereotypes, st);
    read_keyword(st, stringify(UmlDependency));
    read_string_list(relation_default_stereotypes, st);
    read_keyword(st, "end");
    k = read_keyword(st);
  }
  else
    init();
}

bool BrowserPackage::read_stereotypes(const char * f)
{
  char * s = read_file((f == 0) ? "stereotypes" : f);
    
  if (s != 0) {
    try {
      char * st = s;
      char * k = read_keyword(st);
      
      read_stereotypes(st, k);				// updates k
      BrowserClass::read_stereotypes(st, k);		// updates k
      BrowserUseCase::read_stereotypes(st, k);		// updates k
      BrowserArtifact::read_stereotypes(st, k);		// updates k
      BrowserAttribute::read_stereotypes(st, k);	// updates k
      BrowserOperation::read_stereotypes(st, k);	// updates k
      BrowserState::read_stereotypes(st, k);		// updates k
      BrowserPseudoState::read_stereotypes(st, k);	// updates k
      BrowserStateAction::read_stereotypes(st, k);	// updates k
      BrowserComponent::read_stereotypes(st, k);	// updates k
      BrowserDeploymentNode::read_stereotypes(st, k);	// updates k
      BrowserClassView::read_stereotypes(st, k);	// updates k
      BrowserUseCaseView::read_stereotypes(st, k);	// updates k
      BrowserComponentView::read_stereotypes(st, k);	// updates k
      BrowserDeploymentView::read_stereotypes(st, k);	// updates k
      BrowserDiagram::read_stereotypes(st, k);		// updates k
    }
    catch (int) {
      ;
    }
    delete [] s;
    return TRUE;
  }
  
  return FALSE;
}

bool BrowserPackage::import_stereotypes()
{
  QString fn = QFileDialog::getOpenFileName("stereotypes", "stereotypes");
      
  return (!fn.isEmpty() && read_stereotypes((const char *) fn));
}

// save / restore

void BrowserPackage::save(QTextStream & st, bool, QString &) {
  // saves just its reference for its father
  nl_indent(st);
  st << "package_ref " << get_ident() << " // " << get_name();
}

// saves all (modified) packages

void BrowserPackage::init_save_counter() {
  // does nothing
}

void BrowserPackage::save_all(bool modified_only)
{
  IdIterator<BrowserPackage> itcpt(all);

  must_be_saved_counter = 0;
  already_saved = 0;
  
  while (itcpt.current()) {
    BrowserPackage * pack = itcpt.current();

    if (!pack->deletedp() &&
	(!modified_only ||
	 (pack->is_modified && 
	  (pack->is_imported || !pack->is_read_only))))
      pack->BrowserNode::init_save_counter();
    
    ++itcpt;
  }
  
  if (must_be_saved_counter == 0)
    return;
  
  save_progress = new SaveProgress(must_be_saved_counter);
  
  IdIterator<BrowserPackage> it(all);
  QDir d = BrowserView::get_dir();
  QString warning;
  
  while (it.current()) {
    BrowserPackage * pack = it.current();
    bool prj = (pack == BrowserView::get_project());
    QString fn;
    
    if (prj)
      fn.sprintf("%s.prj", (const char *) pack->name);
    else 
      fn.setNum(it.currentKey());
    
    if (pack->deletedp()) {
      backup(d, fn);
    }
    else if (!modified_only ||
	     (pack->is_modified && 
	      (pack->is_imported || !pack->is_read_only))) {
      backup(d, fn);
      
      QFile fp(d.absFilePath(fn));
      
      if (open_file(fp, IO_WriteOnly) != -1) {
	if (prj)
	  UmlWindow::historic_add(fp.name());
	
	QTextStream st(&fp);
	
	st.setEncoding(QTextStream::Latin1);
	
	// saves the package own data
	
	indent0();
	st << "format " << FILEFORMAT << "\n";
	
	save_string(pack->name, st);
	
	if (!prj)
	  st << " // " << (const char *) pack->full_name();

	indent(+1);
	
	if (prj) {
	  GenerationSettings::save_dirs(st);
	  GenerationSettings::save_descriptions(st);
	  nl_indent(st);
	  save_stereotypes();
	  GenerationSettings::save();
	  Tool::save();
	}
	
	nl_indent(st);
	st << "// class settings";
	pack->class_settings.save(st);
	nl_indent(st);
	st << "//class diagram settings";
	pack->classdiagram_settings.save(st);
	nl_indent(st);
	st << "//use case diagram settings";
	pack->usecasediagram_settings.save(st);
	nl_indent(st);
	st << "//sequence diagram settings";
	pack->sequencediagram_settings.save(st);
	nl_indent(st);
	st << "//collaboration diagram settings";
	pack->collaborationdiagram_settings.save(st);
	nl_indent(st);
	st << "//component diagram settings";
	pack->componentdiagram_settings.save(st);
	nl_indent(st);
	st << "//deployment diagram settings";
	pack->deploymentdiagram_settings.save(st);
	nl_indent(st);
	st << "//state diagram settings";
	pack->statediagram_settings.save(st);
	nl_indent(st);
  
	bool nl = FALSE;
	
	save_color(st, "class_color", pack->class_color, nl);
	save_color(st, "duration_color", pack->duration_color, nl);
	save_color(st, "continuation_color", pack->continuation_color, nl);
	save_color(st, "note_color", pack->note_color, nl);
	save_color(st, "fragment_color", pack->fragment_color, nl);
	save_color(st, "usecase_color", pack->usecase_color, nl);
	save_color(st, "package_color", pack->package_color, nl);
	save_color(st, "component_color", pack->component_color, nl);
	save_color(st, "artifact_color", pack->component_color, nl);
	save_color(st, "deploymentnode_color", pack->deploymentnode_color, nl);
	save_color(st, "state_color", pack->state_color, nl);
	save_color(st, "stateaction_color", pack->stateaction_color, nl);
	
	if (prj) {
	  nl_indent(st);
	  st << "font_size " << NormalFont.pointSize();
	  nl_indent(st);
	  st << "diagram_format "
	    << stringify(UmlWindow::default_format()) << '\n';
	  nl_indent(st);
	  st << "mark_for_import";
	  nl_indent(st);
	}
	
	pack->def->save(st, warning);
	
	if (pack->associated_diagram != 0) {
	  if (pack->associated_diagram->deletedp()) {
	    warning += QString("<p>package <b>") + pack->full_name() +
	      "</b>'s associated diagram <b>" +
		pack->associated_diagram->full_name() + "</b> is deleted\n";
	  }
	  else {
	    nl_indent(st);
	    st << "associated_diagram ";
	    pack->associated_diagram->save(st, TRUE, warning);
	  }
	}

	pack->BrowserNode::save(st);
	
	// saves the package's sub elts
	
	QListViewItem * child = pack->firstChild();
	
	if (child != 0) {
	  for (;;) {
	    if (! ((BrowserNode *) child)->deletedp()) {
	      ((BrowserNode *) child)->save(st, FALSE, warning);
	      child = child->nextSibling();
	      if (child != 0)
		st << '\n';
	      else
		break;
	    }
	    else if ((child = child->nextSibling()) == 0)
	      break;
	  }
	}
	
	st << "\nend\n";
	
	pack->is_imported = pack->is_modified = FALSE;
    
	// for saveAs
	if (!modified_only && !pack->is_api_base())
	  pack->is_read_only = FALSE;
      }
      else
	throw 0;
    }
    ++it;
  }
  
  if (save_progress != 0)
    delete save_progress;
  
  if (!warning.isEmpty())
    warn(warning);
  
  delete_backup(d);

  //#warning si multi user indique fin de sauvegarde, les autres users peuvent alors
  //#warning faire des sauvegardes
}

bool BrowserPackage::must_be_saved()
{
  IdIterator<BrowserPackage> it(all);

  while (it.current()) {
    if (it.current()->is_modified &&
	!it.current()->deletedp() &&
	!it.current()->is_read_only)
      return TRUE;
    
    ++it;
  }
  
  return FALSE;
}

void BrowserPackage::save_session(QTextStream & st) {
  if (show_stereotypes)
    st << "show_stereotypes\n";
  
  QString warning;
  
  if (! marked_list.isEmpty()) {    
    QListIterator<BrowserNode> it(marked_list);
    
    st << "marked\n";
    for (; it.current() != 0; ++it) {
      st << "  ";
      it.current()->save(st, TRUE, warning);
      st << '\n';
    }
    st << "end\n";
  }
  
  BrowserNode * sel = (BrowserNode *) listView()->currentItem();
  
  if ((sel != 0) && !sel->deletedp()) {
    st << "selected ";
    sel->save(st, TRUE, warning);
    st << '\n';
  }
  
  if (isOpen()) {
    QListViewItem * child;
    
    st << "open\n";
    
    for (child = firstChild(); child != 0; child = child->nextSibling()) {
      BrowserNode * ch = ((BrowserNode *) child);
      
      if (!ch->deletedp())
	ch->save_open_list(st);
    }
    
    st << "end\n";
  }

  st << "scroll " << listView()->verticalScrollBar()->value()
     << ' ' << listView()->horizontalScrollBar()->value() << '\n';
}

static void open_it(QListViewItem * bn)
{
  if (bn && !bn->isOpen()) {
    open_it(bn->parent());
    bn->setOpen(TRUE);
  }
}

void BrowserPackage::read_session(char * & st, const char * k) {
  if (!strcmp(k, "diagrams")) {
    BrowserDiagram * active_one = 0;
    
    while (strcmp(k = read_keyword(st), "end")) {
      bool active;
      
      if (!strcmp(k, "active")) {
	active = TRUE;
	k = read_keyword(st);
      }
      else
	active = FALSE;
	
      BrowserDiagram * bn = (BrowserDiagram *) get_it(k, read_id(st));
      
      if (bn != 0) {
	ReadContext context = current_context();
	
	bn->open(FALSE);
	restore_context(context);
	bn->read_session(st);
	
	if (active)
	  active_one = bn;
      }
    }
    if (active_one != 0)
      active_one->open(FALSE);
    k = read_keyword(st);
  }
  
  if (!strcmp(k, "show_stereotypes")) {
    if (!show_stereotypes)
      toggle_show_stereotypes();
    k = read_keyword(st);
  }
  else if (show_stereotypes)
    toggle_show_stereotypes();

  if (!strcmp(k, "marked")) {
    while (strcmp(k = read_keyword(st), "end")) {
      BrowserNode * bn = get_it(k, read_id(st));
      
      if (bn != 0)
	bn->toggle_mark();
    }
    k = read_keyword(st);
  }
  
  if (! strcmp(k, "selected")) {
    k = read_keyword(st);
    
    BrowserNode * bn = get_it(k, read_id(st));
      
    if (bn != 0)
      listView()->setSelected(bn, TRUE);
    k = read_keyword(st);
  }
  
  if (! strcmp(k, "open")) {
    while (strcmp(k = read_keyword(st), "end"))
      open_it(get_it(k, read_id(st)));
    k = read_keyword(st);
  }

  if (strcmp(k, "scroll"))
    wrong_keyword(k, "scroll");
  
  listView()->verticalScrollBar()->setValue(read_unsigned(st));
  listView()->horizontalScrollBar()->setValue(read_unsigned(st));
}

BrowserNode * BrowserPackage::get_it(const char * k, int id)
{
  if (!strcmp(k, "package_ref"))
    return all[id];
  
  BrowserNode * r;
  
  if (((r = BrowserUseCaseView::get_it(k, id)) == 0) && 
      ((r = BrowserClassView::get_it(k, id)) == 0) && 
      ((r = BrowserComponentView::get_it(k, id)) == 0) && 
      ((r = BrowserDeploymentView::get_it(k, id)) == 0))
    r = BrowserSimpleRelation::get_it(k, id);
  
  return r;
}

// id is the old package's ident in case of an import
bool BrowserPackage::load(bool recursive, int id) {
  bool less_than_2 = FALSE;
  QString fn;
  bool prj = (this == BrowserView::get_project()) ||
    (this == BrowserView::get_imported_project());
  
  if (prj)
    fn.sprintf("%s.prj", (const char *) name);
  else
    fn.setNum((id == -1) ? get_ident() : id);
  
  QFile fp((in_import())
	   ? BrowserView::get_import_dir().absFilePath(fn)
	   : BrowserView::get_dir().absFilePath(fn));
  int sz;
  
  if ((sz = open_file(fp, IO_ReadOnly)) != -1) {
    ReadContext context = current_context();
    char * s = new char[sz + 1];
    int offset = 0;
    int nread;
    
    is_read_only = !in_import() && read_only_file() || 
      ((user_id() != 0) && is_api_base());
    is_saveable = in_import() || !read_only_file();
      
    do {
      if ((nread = fp.readBlock(s + offset, sz - offset)) == -1) {
	QMessageBox::critical(0, "Error", 
			      BrowserView::get_dir().absFilePath(fn)
			      + "cannot be read");
	delete [] s;
	throw 0;
      }
      offset += nread;
    } while (offset != sz);
  
    s[sz] = 0;
      
    bool err = FALSE;
    unsigned format = read_file_format();
    
    try {
      char * st = s;
      char * k;
      
      read_keyword(st, "format");
      set_read_file_format(read_unsigned(st));
      
      if (read_file_format() > FILEFORMAT) {
	QMessageBox::critical(0, "Error", 
			      "Your version of BOUML is too old to read this project");
	throw 0;
      }
      
      if (in_import() && prj) {
	if (read_file_format() <= 20) {
	  QMessageBox::critical(0, "Error", 
				"\
Sorry, the imported project has a too old format.\n\
To change its format : load this project and save it. The project may be imported.");
	  throw 0;
	}
	
	k = skip_until(st, "mark_for_import");
      }
      else {      
	set_name(read_string(st));
	
	k = read_keyword(st);
	
	if (prj) {
	  less_than_2 = (read_file_format() < 20);
	  GenerationSettings::read_dirs(st, k);		// updates k
	  GenerationSettings::read_descriptions(st, k);		// updates k
	  if (read_file_format() <= 6)
	    GenerationSettings::read(st, k);		// updates k
	  if (read_file_format() <= 7) {
	    read_stereotypes(st, k);			// updates k
	    BrowserClass::read_stereotypes(st, k);	// updates k
	    BrowserUseCase::read_stereotypes(st, k);	// updates k
	    BrowserAttribute::read_stereotypes(st, k);	// updates k
	    BrowserOperation::read_stereotypes(st, k);	// updates k
	    BrowserComponent::read_stereotypes(st, k);	// updates k
	    BrowserDeploymentNode::read_stereotypes(st, k); // updates k
	    BrowserClassView::read_stereotypes(st, k);	// updates k
	    BrowserUseCaseView::read_stereotypes(st, k);// updates k
	    BrowserComponentView::read_stereotypes(st, k); // updates k
	    BrowserDeploymentView::read_stereotypes(st, k); // updates k
	    BrowserDiagram::read_stereotypes(st, k);	// updates k
	    Tool::read(st, k);				// updates k
	  }
	}
	
	if (read_file_format() <= 9)
	  def->read(st, k);			// updates k
	
	class_settings.read(st, k);		// updates k
	classdiagram_settings.read(st, k);	// updates k
	usecasediagram_settings.read(st, k);	// updates k
	sequencediagram_settings.read(st, k);	// updates k
	collaborationdiagram_settings.read(st, k); // updates k
	componentdiagram_settings.read(st, k);	// updates k
	deploymentdiagram_settings.read(st, k);	// updates k
	if (read_file_format() >= 21)
	  statediagram_settings.read(st, k);	// updates k
	
	if (this == BrowserView::get_project()) {
	  // old version
	  if (classdiagram_settings.auto_label_position == UmlDefaultState)
	    classdiagram_settings.auto_label_position = UmlYes;
	  if (usecasediagram_settings.auto_label_position == UmlDefaultState)
	    usecasediagram_settings.auto_label_position = UmlYes;
	  if (componentdiagram_settings.auto_label_position == UmlDefaultState)
	    componentdiagram_settings.auto_label_position = UmlYes;
	  if (deploymentdiagram_settings.auto_label_position == UmlDefaultState)
	    deploymentdiagram_settings.auto_label_position = UmlYes;
	}
	
	read_color(st, "class_color", class_color, k);	// updates k
	read_color(st, "duration_color", duration_color, k);	// updates k
	read_color(st, "continuation_color", continuation_color, k);	// updates k
	read_color(st, "note_color", note_color, k);	// updates k
	read_color(st, "fragment_color", fragment_color, k);	// updates k
	read_color(st, "usecase_color", usecase_color, k);	// updates k
	read_color(st, "package_color", package_color, k);	// updates k
	read_color(st, "component_color", component_color, k);	// updates k
	read_color(st, "artifact_color", artifact_color, k);	// updates k
	read_color(st, "deploymentnode_color", deploymentnode_color, k);	// updates k
	read_color(st, "state_color", state_color, k);		// updates k
	read_color(st, "stateaction_color", stateaction_color, k);		// updates k
      }
      
      if (read_file_format() > 9) {
	if (prj) {
	  if (!strcmp(k, "font_size")) {
	    resize_font((int) read_unsigned(st));
	    k = read_keyword(st);
	  }
	  else
	    init_font();
	  
	  if (!strcmp(k, "diagram_format")) {
	    UmlWindow::set_default_format(canvas_format(read_keyword(st)));
	    k = read_keyword(st);
	  }
	  else
	    UmlWindow::set_default_format(IsoA4);
	  
	  if (strcmp(k, "mark_for_import"))
	    wrong_keyword(k, "mark_for_import");
	  k = read_keyword(st);
	}
	
	def->read(st, k);			// updates k
      }
      else if (prj)
	init_font();
      
      if (!strcmp(k, "associated_class_diagram")) {
	// old format
	set_associated_diagram(BrowserClassDiagram::read_ref(st, "classdiagram_ref"),
			       TRUE);
	k = read_keyword(st);
      }
      else if (!strcmp(k, "associated_diagram")) {
	set_associated_diagram(BrowserClassDiagram::read_ref(st, read_keyword(st)),
			       TRUE);
	k = read_keyword(st);
      }
      
      BrowserNode::read(st, k);	// updates k
      
      if (strcmp(k, "end")) {
	while (BrowserPackage::read(st, k, this, recursive) ||
	       BrowserUseCaseView::read(st, k, this, recursive) ||
	       BrowserClassView::read(st, k, this, recursive) ||
	       ((read_file_format() >= 20) &&
		BrowserComponentView::read(st, k, this, recursive)) ||
	       BrowserDeploymentView::read(st, k, this, recursive) ||
	       BrowserSimpleRelation::read(st, k, this))
	  k = read_keyword(st);
	
	if (strcmp(k, "end"))
	  wrong_keyword(k, "end");
      }
    }
    catch (int) {
      err = TRUE;
    }
    
    delete [] s;
    restore_context(context);
    
    if (!err && (this == BrowserView::get_project())) {
      Tool::read();
      if (read_file_format() > 6) {
	GenerationSettings::read();
	if (read_file_format() > 7)
	  read_stereotypes();
      }
      GenerationSettings::read_includes_imports();
    }
    
    set_read_file_format(format);
    
    if (err)
      throw 0;
  }
  else
    throw 0;
  
  is_imported = is_modified = in_import();
  
  return less_than_2;
}

BrowserPackage * BrowserPackage::read_ref(char * & st)
{
  read_keyword(st, "package_ref");
  
  int id = read_id(st);
  BrowserPackage * result = all[id];
  
  return (result == 0)
    ? new BrowserPackage(id)
    : result;
}

BrowserPackage * BrowserPackage::read(char * & st, char * k,
				      BrowserNode * parent,
				      bool recursive)
{
  if (!strcmp(k, "package_ref")) {
    int id = read_id(st);
    
    BrowserPackage * r = all[id];
    
    if (r == 0)
      r = (parent) 
	? new BrowserPackage("<not yet loaded>", parent, id)
	: new BrowserPackage(id);
    else if (parent != 0) {
      r->set_parent(parent);
      r->set_name("<not yet loaded>");
    }
      
    if (recursive)
      r->load(recursive, id);
      
    return r;
  }
  else
    return 0;
}

void BrowserPackage::package_modified() {
  repaint();
  is_modified = TRUE;
}

void BrowserPackage::renumber(int phase) {
  switch (phase) {
  case -1:
    {
      BrowserNode::renumber(phase);
      
      QDir dir = BrowserView::get_dir();
      const QFileInfoList * l = dir.entryInfoList();
      
      if (l) {
	QListIterator<QFileInfo> it(*l);
	QFileInfo *fi;
	
	while ((fi = it.current()) != 0) {
	  if (fi->isFile() && fi->baseName().at(0).isDigit())
	    dir.rename(fi->fileName(), "_" + fi->fileName());
	  ++it;
	}
      }
      clear(FALSE);
      set_user_id(0);
    }
    return;
  default:
    {
      QDir dir = BrowserView::get_dir();
      QString old = "_" + QString::number(get_ident());
      
      new_ident(phase, all);
      dir.rename(old, QString::number(get_ident()));
      BrowserNode::renumber(phase);
    }
    return;
  case 2:
    {
      QDir dir = BrowserView::get_dir();
      QString filter = "_*";
      const QFileInfoList * l = dir.entryInfoList(filter);
      
      if (l) {
	QListIterator<QFileInfo> it(*l);
	QFileInfo *fi;
	
	while ((fi = it.current()) != 0) {
	  QFile::remove(fi->absFilePath());
	  ++it;
	}
      }
    }
  }
}
