// *************************************************************************
//
// Copyright (C) 2004,2005 Bruno PAGES  All rights reserved.
//
// This file is part of the BOUML Uml Toolkit.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
// e-mail : bouml@free.fr
// home   : http://bouml.free.fr
//
// *************************************************************************

#include <math.h>
#include <qpopupmenu.h> 
#include <qcursor.h>
#include <qpainter.h>

#include "ArrowCanvas.h"
#include "ArrowPointCanvas.h"
#include "ArrowJunctionCanvas.h"
#include "UmlCanvas.h"
#include "LabelCanvas.h"
#include "CdClassCanvas.h"
#include "BrowserDiagram.h"
#include "StereotypeDialog.h"
#include "myio.h"
#include "MenuTitle.h"

ArrowCanvas::ArrowCanvas(UmlCanvas * canvas, DiagramItem * b,
			 DiagramItem * e, UmlCode t, int id)
      : QCanvasPolygon(canvas), DiagramItem(id, canvas), begin(b), end(e),
        itstype(t), label(0), stereotype(0) {
  double bz = begin->get_z();
  double ez = end->get_z();
  
  setZ(((bz < ez) ? bz : ez) - 1);
  b->add_line(this);
  e->add_line(this);
  
  // the first time the canvas must not have a label nor stereotype  
  auto_pos = canvas->browser_diagram()
    ->get_auto_label_position(canvas->browser_diagram()->get_type());
  update_pos();
  
  connect(DrawingSettings::instance(), SIGNAL(changed()),
	  this, SLOT(drawing_settings_modified()));
}

ArrowCanvas::~ArrowCanvas() {
}

void ArrowCanvas::delete_it() {
  if (begin != 0) {
    begin->remove_line(this);
    if (begin->type() == UmlArrowPoint)
      begin->delete_it();
  }
  
  if (end != 0) {
    end->remove_line(this);
    if (end->type() == UmlArrowPoint)
      end->delete_it();
  }
  
  if (label != 0)
    the_canvas()->del(label);
  
  if (stereotype != 0)
    the_canvas()->del(stereotype);
  
  while (!lines.isEmpty())
    lines.first()->delete_it();	// will remove the line
  
  the_canvas()->del(this);
  
  disconnect();
}

UmlCanvas * ArrowCanvas::the_canvas() const {
  return ((UmlCanvas *) canvas());
}

void ArrowCanvas::update_pos() {
  hide_lines();
  
  QPoint old_beginp = beginp;
  QPoint old_endp = endp;
  
  // calcul beginp & endp pour ne plus etre dans b_rct & e_rct
  if (begin->type() == UmlArrowPoint)
    beginp = begin->center();
  else
    begin->shift(beginp, end->center());
  
  switch (end->type()) {
  case UmlArrowPoint:
  case UmlArrowJunction:
    endp = end->center();
    break;
  default:
    end->shift(endp, beginp);
  }
  
  // calcul rectangle englobant
  
  const int dx = endp.x() - beginp.x();
  const int dy = beginp.y() - endp.y();
  
  if ((dx == 0) && (dy == 0))
    return;
  
  const double m = ARROW_LENGTH/sqrt(dx*dx + dy*dy);
  
  double deltax = dy*m;
  double deltay = dx*m;
  
  // modif de delta?/1 pour placer une marge de l/1 a chaque extremite
  boundings.resize(4);
  boundings.setPoint(0, (int) (beginp.x() - deltax - deltay/1),
		     (int) (beginp.y() - deltay + deltax/1));
  boundings.setPoint(1, (int) (beginp.x() + deltax - deltay/1),
		     (int) (beginp.y() + deltay + deltax/1));
  boundings.setPoint(2, (int) (endp.x() + deltax + deltay/1),
		     (int) (endp.y() + deltay - deltax/1));
  boundings.setPoint(3, (int) (endp.x() - deltax + deltay/1),
		     (int) (endp.y() - deltay - deltax/1));
  
  setPoints(boundings);
  
  // points de la fleche
  // *3/5 pour diminuer la fleche et ne pas risquer de sortir de boundings
  deltax *= 3.0/5.0;
  deltay *= 3.0/5.0;
    
  switch (end->type()) {
  case UmlArrowPoint:
    break;
  case UmlArrowJunction:
    // note : the circle or arc is drawn by the arrow junction
    if (itstype == UmlRequired) {
      // end of line
      arrow[0].setX((int) (endp.x() - deltay*(5.0/3.0/ARROW_LENGTH*ARROW_JUNCTION_SIZE/2)));
      arrow[0].setY((int) (endp.y() + deltax*(5.0/3.0/ARROW_LENGTH*ARROW_JUNCTION_SIZE/2)));
      // start angle*16
      arrow[1].setX((dx == 0)
		    ? ((dy > 0) ? 180*16 : 0*16 )
		    : ((((int) (atan(float(dy)/dx)*(180/3.1415927*16)))+
			((dx > 0) ? 90*16 : 270*16))
		       % (360*16)));
      // arc len*16
      arrow[1].setY(180*16);
    }
    else {
      // end of line
      arrow[0].setX((int) (endp.x() - deltay*(5.0/3.0/ARROW_LENGTH*PROVIDED_RADIUS)));
      arrow[0].setY((int) (endp.y() + deltax*(5.0/3.0/ARROW_LENGTH*PROVIDED_RADIUS)));
    }
    break;
  default:
    arrow[0].setX((int) (endp.x() - deltax - deltay));
    arrow[0].setY((int) (endp.y() - deltay + deltax));
    arrow[1].setX((int) (endp.x() + deltax - deltay));
    arrow[1].setY((int) (endp.y() + deltay + deltax));
    arrow[2].setX((int) (endp.x() - deltay));	// pour generalize
    arrow[2].setY((int) (endp.y() + deltax));
  }
  
  if (begin->type() != UmlArrowPoint) {  
    // aggregation's polygone
    switch (itstype) {
    case UmlAggregation:
    case UmlDirectionalAggregation:
    case UmlAggregationByValue:
    case UmlDirectionalAggregationByValue:
      {
	double d_x = (deltax + deltay)*(4.0/5.0);
	double d_y = (deltay - deltax)*(4.0/5.0);
	
	poly.resize(4);
	poly.setPoint(0, beginp.x(), beginp.y());
	poly.setPoint(1, (int) (beginp.x() + d_y), (int) (beginp.y() - d_x));
	poly.setPoint(2, (int) (beginp.x() + d_y + d_x), (int) (beginp.y() - d_x + d_y));
	poly.setPoint(3, (int) (beginp.x() + d_x), (int) (beginp.y() + d_y));
      }
      break;
    default:	// to avoid compiler warning
      break;
    }
  }
  
  if (auto_pos) {
    // moves label dependent on old position
    if ((label != 0) && !label->selected())
      label->moveBy((beginp.x() - old_beginp.x() + endp.x() - old_endp.x())/2.0,
		    (beginp.y() - old_beginp.y() + endp.y() - old_endp.y())/2.0);
    
    // moves stereotype dependent on old position
    if ((stereotype != 0) && !stereotype->selected())
      stereotype->moveBy((beginp.x() - old_beginp.x() + endp.x() - old_endp.x())/2.0,
			 (beginp.y() - old_beginp.y() + endp.y() - old_endp.y())/2.0);
  }
  
  update_show_lines();
}

// force all the arrow points to be ouside r
void ArrowCanvas::move_outside(QRect r) {
  QPoint c = r.center();
  ArrowCanvas * ar = this;
  
  while (ar->end->type() == UmlArrowPoint) {
    ArrowPointCanvas * p = (ArrowPointCanvas *) ar->end;
    QPoint pc = p->center();
    
    int dx = (pc.x() < c.x()) ? -10 : 10;
    int dy = (pc.y() < c.y()) ? -10 : 10;
    
    while (r.contains(pc)) {
      pc.setX(pc.x() + dx);
      pc.setY(pc.y() + dy);
    }
    
    p->move(pc.x(), pc.y());
    
    ar = p->get_other(ar);
  }
}

void ArrowCanvas::change_scale() {
  // all the extremities and the label&stereotype are already moved
  LabelCanvas * lab = label;
  LabelCanvas * ste = stereotype;
  
  label = 0;
  stereotype = 0;
  QCanvasPolygon::setVisible(FALSE);
  update_pos();
  QCanvasPolygon::setVisible(TRUE);
  label = lab;
  stereotype = ste;
}

void ArrowCanvas::setVisible(bool yes) {
  QCanvasPolygon::setVisible(yes);
  if (label != 0)
    label->setVisible(yes);
  if (stereotype != 0)
    stereotype->setVisible(yes);
}

void ArrowCanvas::moveBy(double dx, double dy) {
  if ((label != 0) && !label->selected())
    label->moveBy(dx, dy);
  if ((stereotype != 0) && !stereotype->selected())
    stereotype->moveBy(dx, dy);
}

double ArrowCanvas::get_z() const {
  return z();
}

void ArrowCanvas::set_z(double z) {
  setZ(z);
  if (stereotype != 0)
    stereotype->setZ(z);
}

bool ArrowCanvas::copyable() const {
  return selected() && get_start()->copyable() && get_end()->copyable();    
}

void ArrowCanvas::drawShape(QPainter & p) {
  if (! visible()) return;
  
  p.setBackgroundMode(QObject::OpaqueMode);
  
  switch (itstype) {
  case UmlDirectionalAggregation:
  case UmlDirectionalAggregationByValue:
  case UmlDirectionalAssociation:
  case UmlContain:
  case UmlTransition:
    if (end->type() != UmlArrowPoint) {
      p.drawLine(endp, arrow[0]);
      p.drawLine(endp, arrow[1]);
    }
    // no break
  case UmlAssociation:
  case UmlAggregation:
  case UmlAggregationByValue:
  case UmlLink:
    p.drawLine(beginp, endp);
    if (begin->type() != UmlArrowPoint) {
      switch (itstype) {
      case UmlAggregation:
      case UmlDirectionalAggregation:
	{
	  QBrush brsh = p.brush();
	  
	  p.setBrush(white);
	  p.drawPolygon(poly/*, TRUE*/);
	  p.setBrush(brsh);
	}
	break;
      case UmlAggregationByValue:
      case UmlDirectionalAggregationByValue:
	{
	  QBrush brsh = p.brush();
	  
	  p.setBrush(black);
	  p.drawPolygon(poly/*, TRUE*/);
	  p.setBrush(brsh);
	}      
	break;
      default:	// to avoid compiler warning
	break;
      }
    }
    break;
  case UmlDependency:
  case UmlDependOn:
    if (end->type() != UmlArrowPoint) {
      p.drawLine(endp, arrow[0]);
      p.drawLine(endp, arrow[1]);
    }
    p.setPen(QObject::DotLine);
    p.drawLine(beginp, endp);
    p.setPen(QObject::SolidLine);
    break;
  case UmlRealize:
    p.setPen(QObject::DotLine);
    // no break
  case UmlGeneralisation:
  case UmlInherit:
    if (end->type() != UmlArrowPoint) {
      p.drawLine(beginp, arrow[2]);
      p.setPen(QObject::SolidLine);
      p.drawLine(endp, arrow[0]);
      p.drawLine(endp, arrow[1]);
      p.drawLine(arrow[0], arrow[1]);
    }
    else {
      p.drawLine(beginp, endp);
      p.setPen(QObject::SolidLine);
    }
    break;
  case UmlAnchor:
    p.setPen(QObject::DotLine);
    p.drawLine(beginp, endp);
    p.setPen(QObject::SolidLine);
    break;
  case UmlRequired:
  case UmlProvided:
    if (end->type() != UmlArrowPoint)
      p.drawLine(beginp, arrow[0]);
    else
      p.drawLine(beginp, endp);
    break;
  default:	// to avoid compiler warning
    break;
  }
  
  if (selected()) {
    p.fillRect(beginp.x() - SELECT_SQUARE_SIZE/2,
	       beginp.y() - SELECT_SQUARE_SIZE/2,
	       SELECT_SQUARE_SIZE, SELECT_SQUARE_SIZE, black);
    p.fillRect(endp.x() - SELECT_SQUARE_SIZE/2,
	       endp.y() - SELECT_SQUARE_SIZE/2,
	       SELECT_SQUARE_SIZE, SELECT_SQUARE_SIZE, black);
  }
}

UmlCode ArrowCanvas::type() const {
  return itstype;
}

int ArrowCanvas::rtti() const {
  return RTTI_ARROW;
}

void ArrowCanvas::delete_available(bool &, bool & out_model) const {
  out_model |= TRUE;
}

QPoint ArrowCanvas::center() const {
  return QPoint((beginp.x() + endp.x())/2,
		(beginp.y() + endp.y())/2);
}

bool ArrowCanvas::contains(int, int) const {
  return FALSE;
}

const char * ArrowCanvas::may_start(UmlCode & l) const {
  if (itstype == UmlAnchor)
    return "can't add an anchor to an anchor";
  else if (l != UmlAnchor)
    return "a relation can't have a relation";
  else
    return 0;
}

const char * ArrowCanvas::may_connect(UmlCode l, const DiagramItem * dest) const {
  switch (dest->type()) {
  case UmlNote:
  case UmlIcon:
    return (l == UmlAnchor) ? 0 : "illegal";
  default:
    return "illegal";
  }
}

void ArrowCanvas::connexion(UmlCode action, DiagramItem * dest,
			    const QPoint &, const QPoint &) {
  ArrowCanvas * a =
    new ArrowCanvas(the_canvas(), this, dest, action, 0);
  
  a->show();
  a->modified();
}

bool ArrowCanvas::edit(const QStringList & defaults, 
		       ArrowCanvas * plabel, ArrowCanvas * pstereotype) {
  QString la;
  QString st;
  
  if (plabel)
    la = plabel->label->get_name();
  if (pstereotype)
    st = pstereotype->stereotype->get_name();
  
  StereotypeDialog d(defaults, st, la, QCursor::pos());
  
  if (d.exec() == QDialog::Accepted) {
    bool result = FALSE;
    
    if (la.isEmpty()) {
      if (plabel != 0) {
	// removes it
	the_canvas()->del(plabel->label);
	plabel->label = 0;
	result = TRUE;
      }
    }
    else {
      if (plabel == 0) {
	// adds relation's label
	label = new LabelCanvas(la, the_canvas(), 0, 0);
	default_label_position();	    
	label->show();
	result = TRUE;
      }
      else if ((plabel != 0) &&
	       (plabel->label->get_name() != la)) {
	// update name
	plabel->label->set_name(la);
	plabel->default_label_position();
	result = TRUE;
      }
    }
    
    if (st.isEmpty()) {
      if (pstereotype != 0) {
	// removes it
	the_canvas()->del(pstereotype->stereotype);
	pstereotype->stereotype = 0;
	result = TRUE;
      }
    }
    else {
      if (pstereotype == 0) {
	// adds relation's stereotype
	stereotype = new LabelCanvas(st, the_canvas(), 0, 0);
	default_stereotype_position();	    
	stereotype->show();
	result = TRUE;
      }
      else if ((pstereotype != 0) &&
	       (pstereotype->stereotype->get_name() != st)) {
	// update name
	pstereotype->stereotype->set_name(st);
	pstereotype->default_stereotype_position();
	result = TRUE;
      }
    }
    
    return result;
  }
  else
    return FALSE;
}

void ArrowCanvas::open() {
  if (IsaRelation(itstype)) {
    ArrowCanvas * plabel;
    ArrowCanvas * pstereotype;
    
    search_supports(plabel, pstereotype);
    
    BrowserNode * bn = get_start()->get_bn();

    if ((bn != 0) &&
	edit(bn->default_stereotypes(itstype), plabel, pstereotype)) {
      canvas()->update();
      package_modified();
    }
  }
}

void ArrowCanvas::default_label_position() const {
  QPoint c = center();
  QFontMetrics fm(the_canvas()->get_font(UmlNormalFont));
  
  label->move(c.x() - fm.width(label->get_name())/2, c.y() - fm.height());
}

void ArrowCanvas::default_stereotype_position() const {
  QPoint c = center();
  QFontMetrics fm(the_canvas()->get_font(UmlNormalFont));
  
  stereotype->move(c.x() - fm.width(stereotype->get_name())/2, c.y());
}

void ArrowCanvas::menu(const QPoint&) {
  QPopupMenu m(0);
  ArrowCanvas * plabel;
  ArrowCanvas * pstereotype;
    
  search_supports(plabel, pstereotype);
    
  m.insertItem(new MenuTitle(((plabel == 0) ||
			      plabel->label->get_name().isEmpty())
			     ? QString("line") : plabel->label->get_name(),
			     m.font()), -1);
  if (IsaRelation(itstype)) {
    m.insertSeparator();
    m.insertItem("edit",1);
  }
  if (pstereotype || plabel) {
    m.insertSeparator();
    m.insertItem("select stereotype and label", 2);
    m.insertItem("default stereotype and label position", 3);
    if (plabel && (label == 0))
      m.insertItem("attach label to this segment", 4);
    if (pstereotype && (stereotype == 0))
      m.insertItem("attach stereotype to this segment", 5);
  }
  m.insertSeparator();
  m.insertItem("remove from view",6);
  
  switch (m.exec(QCursor::pos())) {
  case 1:
    open();
    return;
  case 2:
    the_canvas()->unselect_all();
    if (plabel != 0)
      the_canvas()->select(plabel->label);
    if (pstereotype)
      the_canvas()->select(pstereotype->stereotype);
    return;
  case 3:
    if (plabel != 0)
      plabel->default_label_position();
    if (pstereotype)
      pstereotype->default_stereotype_position();
    break;
  case 4:
    label = plabel->label;
    plabel->label = 0;
    default_label_position();
    break;
  case 5:
    stereotype = pstereotype->stereotype;
    pstereotype->stereotype = 0;
    default_stereotype_position();
    break;
  case 6:
    delete_it();
    break;
  }
  
  package_modified();
}

DiagramItem * ArrowCanvas::get_start() const {
  const ArrowCanvas * a = this;
  
  while (a->begin->type() == UmlArrowPoint)
    a = ((ArrowPointCanvas *) a->begin)->get_other(a);
  
  return a->begin;
}

DiagramItem * ArrowCanvas::get_end() const {
  const ArrowCanvas * a = this;
  
  while (a->end->type() == UmlArrowPoint)
    a = ((ArrowPointCanvas *) a->end)->get_other(a);
  
  return a->end;
}

// search label & stereotype supports
void ArrowCanvas::search_supports(ArrowCanvas *& plabel, 
				  ArrowCanvas *& pstereotype) {
  plabel = 0;
  pstereotype = 0;
  
  ArrowCanvas * p;
  
  p = this;
  while (p->end->type() == UmlArrowPoint) {
    if (p->label != 0) plabel = p;
    if (p->stereotype != 0) pstereotype = p;
    p = ((ArrowPointCanvas *) p->end)->get_other(p);
  }

  if (p->label != 0) plabel = p;
  if (p->stereotype != 0) pstereotype = p;

  p = this;
  while (p->begin->type() == UmlArrowPoint) {
    if (p->label != 0) plabel = p;
    if (p->stereotype != 0) pstereotype = p;
    p = ((ArrowPointCanvas *) p->begin)->get_other(p);
  }
  
  if (p->label != 0) plabel = p;
  if (p->stereotype != 0) pstereotype = p;
}

ArrowPointCanvas * ArrowCanvas::brk(const QPoint & p) {
  ArrowPointCanvas * ap =
    new ArrowPointCanvas(the_canvas(), p.x(), p.y());
  
  ap->setZ(z() + 1);
  
  ArrowCanvas * other =
    new ArrowCanvas(the_canvas(), ap, end, itstype, 0);

  ap->add_line(this);
  end->remove_line(this);
  end = ap;
  
  if ((p - beginp).manhattanLength() < (p - endp).manhattanLength()) {
    if (label != 0) {
      other->label = label;
      label = 0;
    }
    
    if (stereotype != 0) {
      other->stereotype = stereotype;
      stereotype = 0;
    }
  }
  
  ap->show();
  other->show();
  
  modified();
  other->modified();
  
  return ap;
}

bool ArrowCanvas::may_join(const ArrowCanvas * other, 
			   const ArrowPointCanvas * ap) const {
  return (end == ((const DiagramItem *) ap))
    ? (begin != other->end)
    : (other->begin != end);
}

ArrowCanvas * ArrowCanvas::join(ArrowCanvas * other, ArrowPointCanvas * ap) {
  // ap is the removed arrow point
  if (end == ap) {
    if (begin == other->end)
      // does not remove the last point of a line linking an element to itself
      return 0;
    
    end = other->end;
    end->add_line(this);	// add before remove in case end is
    end->remove_line(other);	// an arrow junction and not del it
    
    QList<ArrowCanvas> olines = other->lines;
    LabelCanvas * olabel = other->label;
    LabelCanvas * ostereotype = other->stereotype;
    
    other->lines.clear();
    other->begin = other->end = 0;
    other->label = 0;
    other->stereotype = 0;
    ap->remove_line(this);
    ap->remove_line(other);
    the_canvas()->del(other);
    the_canvas()->del(ap);
    hide();
    update_pos();
    show();
    
    //gets other's lines (anchor to note)
    while (! olines.isEmpty()) {
      ArrowCanvas * l = olines.take();
      
      l->hide();
      if (l->begin == other)
	l->begin = this;
      else
	l->end = this;
      add_line(l);
      l->update_pos();
      l->show();
    }
    
    //gets label and stereotype
    QFontMetrics fm(the_canvas()->get_font(UmlNormalFont));
    int h = fm.height();
    QPoint c = center();

    if (olabel) {
      ///tant que gere pas label sur n'importe qu el segment
      if (label != 0)
	label->hide();
      //tq
	
      label = olabel;
      label->move(c.x() - fm.width(label->get_name())/2, c.y() - h);
    }
    if (ostereotype) {
      stereotype = ostereotype;
      stereotype->move(c.x() - fm.width(stereotype->get_name())/2, c.y());
    }
    
    return this;
  }
  else 
    return other->ArrowCanvas::join(this, ap);
}

void ArrowCanvas::select_associated() {
  if (! selected()) {
    the_canvas()->select(this);
    if ((label != 0) && !label->selected())
      the_canvas()->select(label);
    if ((stereotype != 0) && !stereotype->selected())
      the_canvas()->select(stereotype);
    DiagramItem::select_associated();
    begin->select_associated();
    end->select_associated();
  }
}

void ArrowCanvas::modified() {
  if (visible()) {
    hide();
    update_pos();
    show();
    canvas()->update();
    package_modified();
  }
}

void ArrowCanvas::drawing_settings_modified() {
  auto_pos = the_canvas()->browser_diagram()
    ->get_auto_label_position(the_canvas()->browser_diagram()->get_type());
}

void ArrowCanvas::save(QTextStream & st, bool ref, QString & warning) const {
  if (ref)
    st << "line_ref " << get_ident();
  else if (begin->type() != UmlArrowPoint) {
    nl_indent(st);
    st << "line " << get_ident() << ' ' << stringify(itstype);
    indent(+1);
    save_lines(st, TRUE, TRUE, warning);
    indent(-1);
  }
}

const ArrowCanvas * ArrowCanvas::save_lines(QTextStream & st, bool with_label,
					    bool with_stereotype,
					    QString & warning) const {
  nl_indent(st);
  st << "from ref " << begin->get_ident();
  
  const ArrowCanvas * ar = this;
  
  while (ar->end->type() == UmlArrowPoint) {
    st << " z " << z();
    if (with_label && ar->label) {
      st << ' ';
      ar->label->save(st, FALSE, warning);
    }
    if (with_stereotype && ar->stereotype) {
      st << " stereotype ";
      save_string(ar->stereotype->get_name(), st);
      save_xyz(st, ar->stereotype, " xyz");
    }
    st << " to ";
    ar->end->save(st, FALSE, warning);
    ar = ((ArrowPointCanvas *) ar->end)->get_other(ar);
    nl_indent(st);
    st << "line " << ar->get_ident();
  }
  
  st << " z " << z();
  if (with_label && ar->label) {
    st << ' ';
    ar->label->save(st, FALSE, warning);
  }
  if (with_stereotype && ar->stereotype) {
    st << " stereotype ";
    save_string(ar->stereotype->get_name(), st);
    save_xyz(st, ar->stereotype, " xyz");
  }
  st << " to ref " << ar->end->get_ident();
  
  return ar;
}

static ArrowCanvas * make(UmlCanvas * canvas, DiagramItem * b,
			  DiagramItem * e, UmlCode t, int id)
{
  return new ArrowCanvas(canvas, b, e, t, id);
}

ArrowCanvas * ArrowCanvas::read_list(char * & st, UmlCanvas * canvas,
				     UmlCode t, int id,
				     ArrowCanvas * (*pf)(UmlCanvas * canvas, DiagramItem * b,
							 DiagramItem * e, UmlCode t, int id))
{
  char * k;
    
  read_keyword(st, "from");
  read_keyword(st, "ref");
  
  DiagramItem * bi = dict_get(read_id(st), "", canvas);
  ArrowCanvas * result;
  LabelCanvas * label;
  LabelCanvas * stereotype;
  
  for (;;) {
    double z;
    
    if (read_file_format() < 5)
      z = OLD_ARROW_Z;
    else {
      read_keyword(st, "z");
      z = read_double(st);
    }
      
    k = read_keyword(st);
    
    if ((label = LabelCanvas::read(st, canvas, k)) != 0) {
      label->setZ(z);
      k = read_keyword(st);
    }
    if (!strcmp(k, "stereotype")) {
      k = read_string(st);
      if (read_file_format() < 5)
	read_keyword(st, "xy");
      else
	read_keyword(st, "xyz");
      
      int x = (int) read_double(st);
      
      stereotype =
	new LabelCanvas(k, canvas, x, (int) read_double(st));
	
      stereotype->setZ((read_file_format() < 5) ? OLD_ARROW_Z 
						: read_double(st));
      k = read_keyword(st);
    }
    else
      stereotype = 0;
    
    if (strcmp(k, "to"))
      wrong_keyword(k, "to");
    
    DiagramItem * di;
    
    if (!strcmp(k = read_keyword(st), "ref"))
      di = dict_get(read_id(st), "", canvas);
    else {
      di = ArrowPointCanvas::read(st, canvas, k);
      
      if (di == 0)
	unknown_keyword(k);
    }
    
    result = pf(canvas, bi, di, t, id);
    result->setZ(z);
    result->show();
    if (label != 0)
      (result->label = label)->show();
    if (stereotype!= 0)
      (result->stereotype = stereotype)->show();
    
    if (di->type() != UmlArrowPoint)
      return result;
    
    bi = di;
    
    read_keyword(st, "line");
    id = read_id(st);
  }
}

ArrowCanvas * ArrowCanvas::read(char * & st, UmlCanvas * canvas, char * k)
{
  if (!strcmp(k, "line_ref"))
    return ((ArrowCanvas *) dict_get(read_id(st), "arrow canvas", canvas));
  else if (!strcmp(k, "line")) {
    int id = read_id(st);
    UmlCode t = arrow_type(read_keyword(st));
    
    return read_list(st, canvas, t, id, &make);
  }
  else
    return 0;
}

void ArrowCanvas::package_modified() const {
  the_canvas()->browser_diagram()->package_modified();
}

