// *************************************************************************
//
// Copyright (C) 2004,2005 Bruno PAGES  All rights reserved.
//
// This file is part of the BOUML Uml Toolkit.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
// e-mail : bouml@free.fr
// home   : http://bouml.free.fr
//
// *************************************************************************

#include <qpainter.h>
#include <qcursor.h>
#include <qpopupmenu.h> 
#include <qapplication.h>
#include <qclipboard.h>
#include <qpaintdevicemetrics.h>
#include <qfiledialog.h>
#include <qmessagebox.h>
#ifdef WIN32
#include <qtimer.h>
#endif

#include "DiagramView.h"
#include "DiagramWindow.h"
#include "DiagramCanvas.h"
#include "RelationCanvas.h"
#include "SimpleRelationCanvas.h"
#include "TransitionCanvas.h"
#include "CodLinkCanvas.h"
#include "AssocContainCanvas.h"
#include "ArrowPointCanvas.h"
#include "LabelCanvas.h"
#include "NoteCanvas.h"
#include "TextCanvas.h"
#include "BrowserPackage.h"
#include "PackageCanvas.h"
#include "FragmentCanvas.h"
#include "SelectAreaCanvas.h"
#include "BrowserDiagram.h"
#include "UmlWindow.h"
#include "MenuTitle.h"
#include "myio.h"

#include "align_top.xpm"
#include "align_bottom.xpm"
#include "align_left.xpm"
#include "align_right.xpm"
#include "align_center.xpm"
#include "align_vcenter.xpm"
#include "align_hcenter.xpm"

QCString DiagramView::clipboard;
UmlCode DiagramView::copied_from;

DiagramView::DiagramView(QWidget * parent, UmlCanvas * canvas, int i)
    : QCanvasView(canvas, parent), id(i), selectArea(0), start(0),
      line(0), draw_line(FALSE), arrowBeginning(0), do_resize(NoCorner),
      preferred_zoom(0) {
  // enableClipper(TRUE); => probleme d'affichage
  setStaticBackground(TRUE);
  setAcceptDrops(TRUE);
  setDragAutoScroll(TRUE);
  setVScrollBarMode(Auto);
  setHScrollBarMode(Auto);
  setFocusPolicy(QWidget::StrongFocus);
  //setFocusPolicy(QWidget::WheelFocus);
  canvas->setBackgroundColor(QObject::white);
  canvas->setDoubleBuffering(TRUE);
  
  canvas->set_view(this);
}

void DiagramView::init()
{
  clipboard= 0;
}

UmlCanvas * DiagramView::the_canvas() const {
  return (UmlCanvas *) canvas();
}

void DiagramView::contentsMouseDoubleClickEvent(QMouseEvent * e) {
  setFocus();
  unselect_all();
  
  if (draw_line)
    abort_line_construction();
  else {
    QCanvasItem * ci = the_canvas()->collision(e->pos());
    
    if (ci != 0) {
      DiagramItem * di = QCanvasItemToDiagramItem(ci);
      
      if (di != 0)
	di->open();
    }
  }
}

void DiagramView::contentsMousePressEvent(QMouseEvent * e) {
  setFocus();
  if (!window()->frozen()) {
    QCanvasItem * ci = the_canvas()->collision(e->pos());
    
    if (e->button() == QObject::RightButton) {
      // menu on several objects (excluding labels)
  
      if (! draw_line) {
	const QCanvasItemList selected = selection();
	QCanvasItemList::ConstIterator it;
	int n_targets = 0;
	bool in_model = FALSE;
	bool out_model = FALSE;
	bool alignable = FALSE;
	
	for (it = selected.begin(); it != selected.end(); ++it) {
	  if (! isa_label(*it)) {
	    n_targets += 1;
	    
	    DiagramItem * item = QCanvasItemToDiagramItem(*it);
	    
	    item->delete_available(in_model, out_model);
	    alignable |= item->alignable();
	  }	  
	}
	
	if (n_targets > 1) {
	  multiple_selection_menu(in_model, out_model, alignable);
	  return;
	}
      }
      
      unselect_all();
      
      if (draw_line)
	abort_line_construction();
      else if (! BrowserNode::popupMenuActive()) {	// Qt bug
	BrowserNode::setPopupMenuActive(TRUE);
	
	if (ci != 0) {
	  DiagramItem * di = QCanvasItemToDiagramItem(ci);
	  
	  if (di != 0)
	    di->menu(e->pos());
	  BrowserNode::setPopupMenuActive(FALSE);
	}
	else  {
	  menu(e->pos());
	  BrowserNode::setPopupMenuActive(FALSE);
	  // here the window & view may be deleted
	  return;
	}
      }
    }
    else if (! draw_line) {
      UmlCode & action = window()->buttonOn();
      
      switch (action) {
      case UmlSelect:
	mousePressPos = e->pos();
	
	if (ci == 0) {
	  // ne designe pas d objet, vide la selection
	  unselect_all();
	}
	else {
	  const QCanvasItemList selected = selection();
	  QCanvasItemList::ConstIterator it = selected.find(ci);
	  
	  if ((e->state() & ControlButton) == 0) {
	    if (it == selected.end()) {
	      // ne designe pas un objet selectionne, vide la selection
	      unselect_all();
	      
	      // l objet est selectionne
	      select(ci);
	    }
	    // sinon on ne fait rien, l utilisateur veut probablement
	    // deplacer les objets selectionnes
	  }
	  else if (it != selected.end())
	    // l objet etait selectionne
	    unselect(ci);
	  else
	    // l objet n etait pas selectionne
	    select(ci);

	  DiagramItem * di = QCanvasItemToDiagramItem(ci);
	  
	  if (di != 0)
	    do_resize = di->on_resize_point(e->pos());
	}
	break;
      case UmlNote:
	{
	  window()->selectOn();
	  
	  NoteCanvas * note = 
	    new NoteCanvas(the_canvas(), e->x(), e->y(), 0);
	  
	  note->show();
	  note->upper();
	  window()->package_modified();
	}
	break;
      case UmlText:
	{
	  window()->selectOn();
	  
	  TextCanvas * text = 
	    new TextCanvas(the_canvas(), e->x(), e->y(), 0);
	  
	  text->show();
	  text->upper();
	  window()->package_modified();
	}
	break;
      case UmlPackage:
	window()->selectOn();
	{
	  BrowserPackage * p = BrowserPackage::get_package();
	  
	  if (p != 0) {
	    PackageCanvas * pk =
	      new PackageCanvas(p, the_canvas(), e->x(), e->y(), 0);
	    
	    pk->show();
	    pk->upper();
	    window()->package_modified();
	  }
	}
	break;
      case UmlFragment:
	{
	  window()->selectOn();
	  
	  FragmentCanvas * f = 
	    new FragmentCanvas(the_canvas(), e->x(), e->y(), 0);
	  
	  f->show();
	  f->upper();
	  window()->package_modified();
	}
	break;
      default:
	// lines
	unselect_all();
	
	if (ci != 0) {
	  DiagramItem * i = QCanvasItemToDiagramItem(ci);
	  
	  if (i != 0) {
	    const char * err = i->may_start(action);
	    
	    if (err != 0) {
	      QMessageBox::critical(0, "Bouml" , err);
	      window()->selectOn();
	    }
	    else {
	      setCursor(pointingHandCursor);
	      arrowBeginning = start = i;
	      draw_line = TRUE;
	      mousePressPos = e->pos();
	    }
	  }
	  else
	    window()->selectOn();
	}
	else
	  window()->selectOn();
      }
    }
    canvas()->update();
  }
}

void DiagramView::contentsMouseReleaseEvent(QMouseEvent * e) {
  if (!window()->frozen()) {
    if (do_resize != NoCorner)
      do_resize = NoCorner;
    else if (line != 0) {
      the_canvas()->del(line);
      line = 0;
      
      QCanvasItem * ci = the_canvas()->collision(e->pos());
      UmlCode action = window()->buttonOn();
      
      if (ci != 0) {
	DiagramItem * i = QCanvasItemToDiagramItem(ci);
	
	if (i != 0) {
	  const char * err = arrowBeginning->may_connect(action, i);
	  
	  if (err != 0) {
	    QMessageBox::critical(0, "Bouml", err);
	    abort_line_construction();
	  }
	  else {
	    start->connexion(action, i, mousePressPos, e->pos());
	    window()->package_modified();
	    temp.clear();
	    draw_line = FALSE;
	    unsetCursor();
	    arrowBeginning->post_connexion(action, i);
	  }
	}
	else
	  abort_line_construction();
      }
      else if (start->may_connect(action)) {
	// component required/provided interface
	if (start->connexion(action, mousePressPos, e->pos())) {
	  window()->package_modified();
	  temp.clear();
	  draw_line = FALSE;
	  unsetCursor();
	}
	else
	  abort_line_construction();
      }
      else if (arrowBeginning->allowed_direction(action) == DiagramItem::All) {
	// adds an ArrowPoint and the line to it
	ArrowPointCanvas * ap =
	  new ArrowPointCanvas(the_canvas(), e->x(), e->y());
	
	ap->upper();
	
	UmlCode t = window()->browser_diagram()->get_type();
	ArrowCanvas * a;
	
	if ((t == UmlClassDiagram) && IsaRelation(action))
	  a = new RelationCanvas(the_canvas(), start, ap, 0, action, 0);
	else if ((t == UmlColDiagram) && (action == UmlLink))
	  a = new CodLinkCanvas(the_canvas(), start, ap, 0);
	else if (((t == UmlComponentDiagram) || (t == UmlDeploymentDiagram))
		 && (action == UmlContain))
	  a = new AssocContainCanvas(the_canvas(), start, ap, 0);
	else if ((t == UmlStateDiagram) && (action == UmlTransition))
	  a = new TransitionCanvas(the_canvas(), start, ap, 0, 0);
	else if (IsaSimpleRelation(action))
	  a = new SimpleRelationCanvas(the_canvas(), start, ap, 0, action, 0);
	else
	  a = new ArrowCanvas(the_canvas(), start, ap, action, 0);
	
	temp.append(a);		// before the point, see abort_line_construction()
	temp.append(ap);
	
	ap->show();
	a->show();
	
	// re create immediatly the line
	draw_line = TRUE;
	mousePressPos = e->pos();
	start = ap;
	line = new QCanvasLine(canvas());
	line->setZ(TOP_Z);
	line->setPoints(e->pos().x(), e->pos().y(), 
			e->pos().x(), e->pos().y());
	line->setPen(QObject::DotLine);
	line->show();
	line->setPen(QObject::SolidLine);
	
	canvas()->update();
	window()->package_modified();
	
	// remark : mouse tracking does not work (?)
	setCursor(pointingHandCursor);
	
	return;
      }
      else
	abort_line_construction();
      
      canvas()->update();
      window()->selectOn();
    }
    else if ((e->button() != QObject::RightButton) && (selectArea != 0)) {
      // selectionne les objets dans la zone de selection
      QRect r = selectArea->boundingRect();
      
      the_canvas()->del(selectArea);
      selectArea = 0;
      
      QCanvasItemList l = canvas()->collisions(r);
      QCanvasItemList::Iterator it;
      
      for (it = l.begin(); it != l.end(); ++it)
	if (r.contains((*it)->boundingRect(), TRUE) && // completement inclus
	    !isa_alien(*it) &&		// DiagramItem
	    (*it)->visible() &&		// pas en attente de destruction
	    !(*it)->selected())   	// pas deja selectionne
	  select(*it);
      
      canvas()->update();
    }
  }
}

void DiagramView::contentsMouseMoveEvent(QMouseEvent * e) {
  if (!window()->frozen()) {
    if (e->button() != QObject::RightButton) {
      int dx = e->pos().x() - mousePressPos.x();
      int dy = e->pos().y() - mousePressPos.y();
      QCanvasItemList selected = selection();
      
      if (! selected.isEmpty()) {
	// deplace/redimentionne les objets selectionnes
	if ((selected.count() == 1) && isa_arrow(selected.first())) {
	  // cuts the line in two adding an ArrowPoint which
	  // becomes the alone selected item allowing to move it
	  ArrowCanvas * a = ((ArrowCanvas *) selection().first());
	  
	  unselect(a);
	  select(a->brk(e->pos()));
	  window()->package_modified();
	}
	else
	  moveSelected(dx, dy);
	
	mousePressPos = e->pos();
      }
      else if (draw_line) {
	if (line == 0) {
	  // premier deplacement : cree la ligne
	  line = new QCanvasLine(canvas());
	  line->setZ(TOP_Z);
	  switch (arrowBeginning->allowed_direction(window()->buttonOn())) {
	  case DiagramItem::Horizontal:
	    // horizontal line
	    line->setPoints(mousePressPos.x(), mousePressPos.y(), 
			    e->pos().x(), mousePressPos.y());
	    break;
	  default:
	    line->setPoints(mousePressPos.x(), mousePressPos.y(), 
			    e->pos().x(), e->pos().y());
	  }
	  line->setPen(QObject::DotLine);
	  line->show();
	  line->setPen(QObject::SolidLine);
	}
	else {
	  QPoint st = line->startPoint();
	  
	  switch (arrowBeginning->allowed_direction(window()->buttonOn())) {
	  case DiagramItem::Horizontal:
	    // horizontal line
	    line->setPoints(st.x(), st.y(), e->pos().x(), st.y());
	    break;
	  default:
	    line->setPoints(st.x(), st.y(), e->pos().x(), e->pos().y());
	  }
	}
      }
      else {
	// la selection est vide, veut probablement la faire en faisant
	// une zone via la souris, montre cette zone
	if (selectArea == 0) {
	  // premier deplacement : cree la zone
	  selectArea = 
	    new SelectAreaCanvas(mousePressPos.x(), mousePressPos.y(), dx, dy, canvas());
	  
	  selectArea->setZ(TOP_Z);
	  selectArea->show();
	}
	else {
	  // redimensionne la zone
	  selectArea->setSize(dx, dy);
	}
      }    
      
      canvas()->update();
    }
  }
}

void DiagramView::abort_line_construction() {
  if (draw_line) {
    draw_line = FALSE;
    unsetCursor();
        
    if (line != 0) {
      the_canvas()->del(line);
      line = 0;
    }
    
    if (! temp.isEmpty()) {
      QCanvasItemList::Iterator it = temp.begin();
      
      // remove the line in the start item
      arrowBeginning->remove_line(((ArrowCanvas *) *it));
      
      do {
	the_canvas()->del(*it);
	it = temp.remove(it);
      } while (it != temp.end());
    }
      
    canvas()->update();
  }
}

#ifdef WIN32
#pragma optimize("", off)
// generated code is wrong if optimize speed
#endif
void DiagramView::delete_them(bool in_model) {
  for (;;) {
    const QCanvasItemList & selected = selection();

    if (selected.isEmpty())
      break;

    QCanvasItemList::ConstIterator it = selected.begin();
    DiagramItem * item = QCanvasItemToDiagramItem(*it);
    bool in = FALSE;
    bool out = FALSE;

    item->delete_available(in, out);
    if (in || out)
      item->remove(in_model & in);		// modify selection()
    else
      the_canvas()->unselect(*it);
  }
  window()->package_modified();
}
#ifdef WIN32
#pragma optimize("", on)
#endif

void DiagramView::alignLeft() {
  const QCanvasItemList selected = selection();
  QCanvasItemList::ConstIterator it = selected.begin();
  double ref = (*it)->x();
  
  while ((++it)  != selected.end())
    (*it)->moveBy(ref - (*it)->x(), 0);
}

void DiagramView::alignRight() {
  const QCanvasItemList selected = selection();
  QCanvasItemList::ConstIterator it = selected.begin();
  double ref = (*it)->boundingRectAdvanced().right();
  
  while ((++it)  != selected.end())
    (*it)->moveBy(ref - (*it)->boundingRectAdvanced().right(), 0);
  
  window()->package_modified();
}

void DiagramView::alignTop() {
  const QCanvasItemList selected = selection();
  QCanvasItemList::ConstIterator it = selected.begin();
  double ref = (*it)->y();
  
  while ((++it)  != selected.end())
    (*it)->moveBy(0, ref - (*it)->y());
}

void DiagramView::alignBottom() {
  const QCanvasItemList selected = selection();
  QCanvasItemList::ConstIterator it = selected.begin();
  double ref = (*it)->boundingRectAdvanced().bottom();
  
  while ((++it)  != selected.end())
    (*it)->moveBy(0, ref - (*it)->boundingRectAdvanced().bottom());
  
  window()->package_modified();
}

void DiagramView::alignHorizontaly() {
  const QCanvasItemList selected = selection();
  QCanvasItemList::ConstIterator it = selected.begin();
  QRect r = (*it)->boundingRectAdvanced();
  double ref = (r.top() + r.bottom())/2;
  
  while ((++it)  != selected.end()) {
    r = (*it)->boundingRectAdvanced();
    
    (*it)->moveBy(0, ref - (r.top() + r.bottom())/2);
  }
  
  window()->package_modified();
}

void DiagramView::alignVerticaly() {
  const QCanvasItemList selected = selection();
  QCanvasItemList::ConstIterator it = selected.begin();
  QRect r = (*it)->boundingRectAdvanced();
  double ref = (r.left() + r.right())/2;
  
  while ((++it)  != selected.end()) {
    r = (*it)->boundingRectAdvanced();
    
    (*it)->moveBy(ref - (r.left() + r.right())/2, 0);
  }
  
  window()->package_modified();
}

void DiagramView::alignCenter() {
  const QCanvasItemList selected = selection();
  QCanvasItemList::ConstIterator it = selected.begin();
  QRect r = (*it)->boundingRectAdvanced();
  double refx = (r.left() + r.right())/2;
  double refy = (r.top() + r.bottom())/2;
  
  while ((++it)  != selected.end()) {
    r = (*it)->boundingRectAdvanced();
    
    (*it)->moveBy(refx - (r.left() + r.right())/2,
		  refy - (r.top() + r.bottom())/2);
  }
  
  window()->package_modified();
}


void DiagramView::multiple_selection_menu(bool in_model, bool out_model,
					  bool alignable) {
  const QCanvasItemList selected = selection();
  QCanvasItemList::ConstIterator it;
  QPopupMenu m(0);
  bool used = FALSE;
  
  m.insertItem(new MenuTitle("Multiple selection menu", m.font()), -1);
  m.insertSeparator();
  for (it = selected.begin(); it != selected.end(); ++it) {
    if (QCanvasItemToDiagramItem(*it)->linked()) {
      m.insertItem("select linked items", 1);
      used = TRUE;
      m.insertSeparator();
      break;
    }
  }
  
  QPopupMenu al(0);
  
  if (alignable) {
    QPixmap top((const char **) align_top);
    QPixmap bottom((const char **) align_bottom);
    QPixmap left((const char **) align_left);
    QPixmap right((const char **) align_right);
    QPixmap center((const char **) align_center);
    QPixmap vcenter((const char **) align_vcenter);
    QPixmap hcenter((const char **) align_hcenter);
    
    al.insertItem(top, "align top", 4);
    al.insertItem(bottom, "align bottom", 5);
    al.insertItem(left, "align left", 6);
    al.insertItem(right, "align right", 7);
    al.insertItem(center, "align center", 8);
    al.insertItem(vcenter, "align center verticaly", 9);
    al.insertItem(hcenter, "align center horizontaly", 10);

    m.insertItem("align", &al);
    m.insertSeparator();
  }
  
  m.insertItem("copy selected (Ctrl+c)", 11);
  m.insertItem("cut selected (Ctrl+x, remove from view)", 12);
  
  if (out_model)
    m.insertItem("remove selected from view (Suppr)", 2);
  if (in_model)
    m.insertItem("delete selected (Control+d)", 3);

  if (used || out_model || in_model || alignable) {
    switch (m.exec(QCursor::pos())) {
    case 1:
      unselect_all();
      for (it = selected.begin(); it != selected.end(); ++it)
	QCanvasItemToDiagramItem(*it)->select_associated();
      break;
    case 12:
      clipboard = window()->copy_selected();
      copied_from = window()->browser_diagram()->get_type();
      // no break !
    case 2:
      delete_them(FALSE);
      break;
    case 3:
      delete_them(TRUE);
      break;
    case 4:
      alignTop();
      break;
    case 5:
      alignBottom();
      break;
    case 6:
      alignLeft();
      break;
    case 7:
      alignRight();
      break;
    case 8:
      alignCenter();
      break;
    case 9:
      alignVerticaly();
      break;
    case 10:
      alignHorizontaly();
      break;
    case 11:
      clipboard = window()->copy_selected();
      copied_from = window()->browser_diagram()->get_type();
      return;
    default:
      return;
    }
    
    canvas()->update();
  }
}

void DiagramView::moveSelected(int dx, int dy) {
  QCanvasItemList selected = selection();
  QCanvasItemList::ConstIterator it;
  
  for (it = selected.begin(); it != selected.end(); ++it)
    QCanvasItemToDiagramItem(*it)
      ->prepare_for_move(do_resize != NoCorner);
  
  selected = selection();
  
  for (it = selected.begin(); it != selected.end(); ++it) {
    DiagramItem * di = QCanvasItemToDiagramItem(*it);
    
    if (do_resize)
      di->resize(do_resize, dx, dy);
    else
      (*it)->moveBy(dx, dy);
  }
  
  window()->package_modified();
}

void DiagramView::keyPressEvent(QKeyEvent * e) {
  if (!window()->frozen()) {
    if (e->state() == QObject::ControlButton) {
      switch (e->key()) {
      case QObject::Key_A:
	select_all();
	break;
      case QObject::Key_C: 
	clipboard = window()->copy_selected();
	copied_from = window()->browser_diagram()->get_type();
	break;
      case QObject::Key_D:
	delete_them(TRUE);
	break;
      case QObject::Key_S:
	UmlWindow::save_it();
	break;
      case QObject::Key_V:
	if (!clipboard.isEmpty() &&
	    (copied_from == window()->browser_diagram()->get_type()))
	  paste();
	break;
      case QObject::Key_X:
	clipboard = window()->copy_selected();
	copied_from = window()->browser_diagram()->get_type();
	delete_them(FALSE);
	break;
      }
    }
    else {
      switch (e->key()) {
      case QObject::Key_Delete:
	delete_them(FALSE);
	break;
      case QObject::Key_Left:
	moveSelected(-1, 0);
	break;
      case QObject::Key_Up:
	moveSelected(0, -1);
	break;
      case QObject::Key_Right:
	moveSelected(1, 0);
	break;
      case QObject::Key_Down:
	moveSelected(0, 1);
	break;
      }
    }
    
    canvas()->update();
  }
  
  e->ignore();
}

void DiagramView::select_all() {
  QCanvasItemList all = canvas()->allItems();
  QCanvasItemList::Iterator cit;

  for (cit = all.begin(); cit != all.end(); ++cit) {
    if ((QCanvasItemToDiagramItem(*cit) != 0) && // an uml canvas item
	!(*cit)->selected() &&
	(*cit)->visible())
      select(*cit);
  }
  
  canvas()->update();
}

void DiagramView::set_zoom(double zoom) {
  ((UmlCanvas *) canvas())->set_zoom(zoom);
  
  QCanvasItemList all = canvas()->allItems();
  QCanvasItemList::Iterator cit;

  // hack to freeze arrow's labels position : select all
  QCanvasItemList selected = selection();
  
  for (cit = all.begin(); cit != all.end(); ++cit) {
    if ((QCanvasItemToDiagramItem(*cit) != 0) && // an uml canvas item
	!(*cit)->selected() &&
	(*cit)->visible())
      select(*cit);
  }
  
  DiagramItem * di;
  
  for (cit = all.begin(); cit != all.end(); ++cit)
    if ((*cit)->visible() && 
	!isa_arrow(*cit) &&
	((di = QCanvasItemToDiagramItem(*cit)) != 0))
      di->change_scale();
  
  for (cit = all.begin(); cit != all.end(); ++cit)
    if ((*cit)->visible() && isa_arrow(*cit))
      ((ArrowCanvas *) *cit)->change_scale();
  
  for (cit = all.begin(); cit != all.end(); ++cit) {
    if ((QCanvasItemToDiagramItem(*cit) != 0) && // an uml canvas item
	!(*cit)->selected() &&
	(*cit)->visible())
      select(*cit);
  }
  
  // hack end
  unselect_all();
  for (cit = selected.begin(); cit != selected.end(); ++cit)
    select(*cit);
  
  ((UmlCanvas *) canvas())->zoom_end();
  
  canvas()->update();
}

void DiagramView::needed_width_height(int & maxx, int & maxy) const {
  // search the max used x and y
  QCanvasItemList all = canvas()->allItems();
  QCanvasItemList::Iterator cit;
  
  maxx = 0;
  maxy = 0;
  
  for (cit = all.begin(); cit != all.end(); ++cit) {
    DiagramItem * di;
    
    if ((*cit)->visible() && 
	((di = QCanvasItemToDiagramItem(*cit)) != 0)) {
      if (di->type() != UmlLifeLine) {
	QRect r = (*cit)->boundingRect();
	
	if (r.right() > maxx)
	  maxx = r.right();
	if (r.bottom() > maxy)
	  maxy = r.bottom();
      }
    }
  }
}

void DiagramView::fit_scale() {
  int maxx;
  int maxy;
  
  needed_width_height(maxx, maxy);
  
  double scale = the_canvas()->zoom() * visibleWidth() / maxx;
  double scale_y = the_canvas()->zoom() * visibleHeight() / maxy;
  
  if (scale > scale_y)
    scale = scale_y;
  
  scale -= 0.02;
  
  set_zoom((scale < SCALE_MIN/100.0) 
	   ? SCALE_MIN/100.0
	   : ((scale > SCALE_MAX/100.0) ? SCALE_MAX/100.0 : scale));
}

void DiagramView::optimal_window_size() {
  int maxx;
  int maxy;
  
  needed_width_height(maxx, maxy);
  
  window()->resize((int) (maxx * 1.02) + window()->width() - visibleWidth(),
		   (int) (maxy * 1.02) + window()->height() - visibleHeight());
}


void DiagramView::preferred_size_zoom() {
  if (preferred_zoom != 0) {
    window()->resize(preferred_size);
    window()->new_scale((int) (preferred_zoom * 100));
  }
}

void DiagramView::set_format(int f) {
  if (f != (int) window()->browser_diagram()->get_format()) {
    window()->browser_diagram()->set_format((CanvasFormat) f);
    the_canvas()->resize((CanvasFormat) f);
  }
}

int DiagramView::default_menu(QPopupMenu & m, int f) {
  QPopupMenu formatm(0);
  
  m.insertItem("edit drawing settings", 0);
  m.insertSeparator();
  m.insertItem("select diagram in browser", 1);
  m.insertItem("select all (Ctrl+a)", 2);
  m.insertSeparator();
  m.insertItem("optimal scale", 7);
  m.insertItem("optimal window size", 8);
  m.insertSeparator();
  m.insertItem("copy visible picture part", 3);
  m.insertItem("save visible picture part", 10);
  if (!clipboard.isEmpty() &&
      (copied_from == window()->browser_diagram()->get_type()))
    m.insertItem("paste copied items (Ctrl+v)", 9);
  m.insertSeparator();
  m.insertItem("set preferred size and scale", 4);
  if (preferred_zoom != 0)
    m.insertItem("restore preferred size and scale", 5);
  init_format_menu(formatm, f);
  m.insertItem("format", &formatm);
  m.insertSeparator();
  m.insertItem("undo changes", 6);

  int choice = m.exec(QCursor::pos());
  
  switch (choice) {
  case 1:
    the_canvas()->browser_diagram()->select_in_browser();
    break;
  case 2:
    select_all();
    break;
  case 3:
    the_canvas()->show_limits(FALSE);
    QApplication::clipboard()->setPixmap(QPixmap::grabWidget(viewport()));
    the_canvas()->show_limits(TRUE);
    break;
  case 4:
    preferred_zoom = the_canvas()->zoom();
    preferred_size.setWidth(window()->width());
    preferred_size.setHeight(window()->height());
    window()->package_modified();
    break;
  case 5:
    preferred_size_zoom();
    break;
  case 7:
    window()->fit_scale();
    break;
  case 8:
    optimal_window_size();
    the_canvas()->update();
    // do it an other time because of the scroll bars
#ifdef WIN32
    QTimer::singleShot(1000, this, SLOT(optimal_window_size()));
#else
    optimal_window_size();
#endif
    break;
  case 9:
    paste();
    break;
  case 10:
    save_picture();
    break;
  default:
    if (choice >= f) {
      set_format(choice - f);
      window()->package_modified();
    }
  }
  
  return choice;
}

void DiagramView::init_format_menu(QPopupMenu & m, int f) const {
  m.setCheckable(TRUE);
  
  for (int i = 0; i != CanvasFormatSup; i += 1)
    m.insertItem(stringify((CanvasFormat) i), f + i);
  
  m.setItemChecked(f + (int) window()->browser_diagram()->get_format(),
		   TRUE);
}

void DiagramView::load(const char * pfix) {
  // for a re-load :
  unselect_all();
  preferred_zoom = 0;
  
  QCanvasItemList all = canvas()->allItems();
  QCanvasItemList::Iterator it;
  DiagramItem * di;
  
  for (it = all.begin(); it != all.end(); ++it)
    if (((di = QCanvasItemToDiagramItem(*it)) != 0) && // an uml canvas item
	(*it)->visible())
      di->delete_it();
  
  ((UmlCanvas *) canvas())->set_zoom(1);
  ((UmlCanvas *) canvas())->zoom_end();
  
  // load
  BrowserNode::pre_load();
  
  try {
    QString t = pfix;
    
    t += " diagram " + ((QWidget *) parent())->caption();
    
    read_in(t);
    read();
  }
  catch (int) {
    ;
  }
  
  BrowserNode::post_load();
  DiagramItem::post_load();
  
  // to managed deleted items present in the browser
  all = canvas()->allItems();  
  
  for (it = all.begin(); it != all.end(); ++it) {
    if ((*it)->visible() &&
	((di = QCanvasItemToDiagramItem(*it)) != 0) && 
	(di->get_bn() != 0) && // do not manage relations here
	di->get_bn()->deletedp())
      di->delete_it();
  }
  
  if (window()->get_view() != 0)
    // re-load
    preferred_size_zoom();
  canvas()->update();
}

void DiagramView::read() {
  char * s;
  
  if ((s = read_definition(id, "diagram")) != 0) {
    unsigned format = read_file_format();
    bool error = FALSE;
    
    try {
      char * st = s;
      char * k;
      
      k = read_keyword(st);
      if (!strcmp(k, "format")) {
	set_read_file_format(read_unsigned(st));
	k = read_keyword(st);
      }
      else
	// format saved in diagram since format 4
	set_read_file_format(3);
      
      the_canvas()->read_old_diagram(read_file_format() < 9);
      
      read(st, k);
    }
    catch (...) {
      error = TRUE;
    }
    
    delete [] s;
    the_canvas()->read_old_diagram(FALSE);
    set_read_file_format(format);
    the_canvas()->clear_old_ids();
    
    if (error)
      throw 0;
  }
}

void DiagramView::paste() {
  double old_zoom = ((UmlCanvas *) canvas())->zoom();
  char * s = new char[clipboard.length() + 1];
  
  memcpy(s, (const char *) clipboard, clipboard.length() + 1);
  
  unselect_all();
  set_zoom(1);
  the_canvas()->set_paste(TRUE);
  read_in("clipboard");
  
  try {
    char * st = s;
    char * k;
      
    k = read_keyword(st);
    if (!strcmp(k, "format")) {
      set_read_file_format(read_unsigned(st));
      k = read_keyword(st);
    }
    else
      // format saved in diagram since format 4
      set_read_file_format(3);
      
    read(st, k);
  }
  catch (int) {
    ;
  }
  
  delete [] s;
  
  the_canvas()->clear_old_ids();
  the_canvas()->set_paste(FALSE);
  DiagramItem::post_load();
  
  // to managed deleted items present in the browser
  // compute rect containing all the paste items
  QCanvasItemList l = selection();
  QCanvasItemList::Iterator it;
  DiagramItem * di;
  double minx = 1e10;
  double maxx = 0;
  double miny = 1e10;
  double maxy = 0;
  
  for (it = l.begin(); it != l.end(); ++it) {
    if ((*it)->visible() &&
	((di = QCanvasItemToDiagramItem(*it)) != 0) && 
	(di->get_bn() != 0) && // do not manage relations here
	di->get_bn()->deletedp())
      di->delete_it();
    else {
      QRect r = (*it)->boundingRect();
      
      if (r.x() < minx)
	minx = r.x();
      if (r.y() < miny)
	miny = r.y();
      if (r.x() > maxx)
	maxx = r.right();
      if (r.y() > maxy)
	maxy = r.bottom();
    }
  }

  // place all paste items in the center of the viewport
  double dx = contentsX() + width()/2 - (minx + maxx)/2;
  double dy = contentsY() + height()/2 - (miny + maxy)/2;
  
  l = selection();
  
  for (it = l.begin(); it != l.end(); ++it)
    QCanvasItemToDiagramItem(*it)->prepare_for_move(FALSE);
  
  l = selection();
  
  for (it = l.begin(); it != l.end(); ++it)
    (*it)->moveBy(dx, dy);
  
  set_zoom(old_zoom);
  canvas()->update();
  window()->package_modified();
}

bool DiagramView::save_in(const char * f) {
  if (the_canvas()->selection().count() != 0) {
    // unselect element and redraw them to remove selection mark
    unselect_all();
    the_canvas()->setAllChanged();
    the_canvas()->update();
  }
  
  the_canvas()->show_limits(FALSE);
  
  bool r = QPixmap::grabWidget(viewport()).save(f, "PNG");
  
  the_canvas()->show_limits(TRUE);
  return r;
}

void DiagramView::save_picture() {
  QString filename =
    QFileDialog::getSaveFileName(QString::null, "PNG file (*.png)", this);

  if (!filename.isNull()) {
    if (filename.right(4).lower() != ".png")
      filename += ".png";
  
    save_in(filename);
  }
}

void DiagramView::print(QPainter * p) {
  if (the_canvas()->selection().count() != 0)
    unselect_all();

  double old_zoom = ((UmlCanvas *) canvas())->zoom();
  
  if (old_zoom < 1)
    set_zoom(1);
  
  QPaintDeviceMetrics m(p->device());
  int devh = m.height();
  int devw = m.width();
  int h = contentsHeight();
  int w = contentsWidth();
  double zoom = ((double) devh)/h;

  if (((double) devw)/w < zoom)
    zoom = ((double) devw)/w;

  zoom  *= 0.95;
  p->scale(zoom, zoom);
  ((UmlCanvas *) canvas())->show_limits(FALSE);
  drawContents(p, 0, 0, w, h);
  ((UmlCanvas *) canvas())->show_limits(TRUE);
  
  if (old_zoom < 1)
    set_zoom(old_zoom);
}

void DiagramView::renumber(int ident) {
  id = ident;
}

void DiagramView::save_session(QTextStream & st) {
  st << (int) (((UmlCanvas *) canvas())->zoom() * 100)
     << ' ' << (int) window()->browser_diagram()->get_format()
     << ' ' << verticalScrollBar()->value()
     << ' ' << horizontalScrollBar()->value() << '\n';
}

void DiagramView::read_session(char * & st) {
  // zoom managed by DiagramWindow
  canvas()->update();
  set_format((int) read_unsigned(st));
  verticalScrollBar()->setValue(read_unsigned(st));
  horizontalScrollBar()->setValue(read_unsigned(st));
  canvas()->update();
}
