// *************************************************************************
//
// Copyright (C) 2004,2005 Bruno PAGES  All rights reserved.
//
// This file is part of the BOUML Uml Toolkit.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
// e-mail : bouml@free.fr
// home   : http://bouml.free.fr
//
// *************************************************************************

#include <math.h>
#include <qpainter.h>

#include "LabelCanvas.h"
#include "DiagramCanvas.h"
#include "myio.h"

LabelCanvas::LabelCanvas(const QString & n, UmlCanvas * canvas, int x, int y,
			 bool bold, bool italic, bool underlined)
    : QCanvasText(n, canvas), DiagramItem(-1, canvas) {
  if (bold)
    setFont((italic) ? ((UmlCanvas *) canvas)->get_font(UmlNormalBoldItalicFont)
		     : ((UmlCanvas *) canvas)->get_font(UmlNormalBoldFont));
  else if (italic)
    setFont(((UmlCanvas *) canvas)->get_font(UmlNormalItalicFont));
  else
    setFont((underlined) ? ((UmlCanvas *) canvas)->get_font(UmlNormalUnderlinedFont)
			 : ((UmlCanvas *) canvas)->get_font(UmlNormalFont));
  setX(x);
  setY(y);
  setZ(LABEL_Z);
  set_center100();  
  if (canvas->paste())
    canvas->select(this);
}

LabelCanvas::~LabelCanvas() {
}

void LabelCanvas::delete_it() {
  ((UmlCanvas *) canvas())->del(this);
}

UmlCanvas * LabelCanvas::the_canvas() const {
  return ((UmlCanvas *) canvas());
}

void LabelCanvas::set_name(const QString & s) {
  hide();
  
  setText(s);
  if (font().strikeOut())
    setFont(((UmlCanvas *) canvas())->get_font(UmlNormalFont));

  show();
  canvas()->update();
}

void LabelCanvas::set_strikeout(bool yes) {
  if (yes) {
    if (!font().strikeOut())
      setFont(((UmlCanvas *) canvas())->get_font(UmlNormalStrikeOutFont));
  }
  else if (font().strikeOut())
    setFont(((UmlCanvas *) canvas())->get_font(UmlNormalFont));
}

void LabelCanvas::change_scale() {
  hide();

  if (font().bold())
    setFont((font().italic()) ? ((UmlCanvas *) canvas())->get_font(UmlNormalBoldItalicFont)
			      : ((UmlCanvas *) canvas())->get_font(UmlNormalBoldFont));
  else if (font().italic())
    setFont(((UmlCanvas *) canvas())->get_font(UmlNormalItalicFont));
  else if (font().underline())
    setFont(((UmlCanvas *) canvas())->get_font(UmlNormalUnderlinedFont));
  else
    setFont((font().strikeOut()) ? ((UmlCanvas *) canvas())->get_font(UmlNormalStrikeOutFont)
				 : ((UmlCanvas *) canvas())->get_font(UmlNormalFont));
  recenter();
  
  show();
}

void LabelCanvas::recenter() {
  double scale = the_canvas()->zoom();
  QPoint c = center();
  
  QCanvasText::moveBy(center_x_scale100 * scale - c.x(),
		      center_y_scale100 * scale - c.y());
}

void LabelCanvas::set_center100() {
  QPoint c = center();
  double scale = the_canvas()->zoom();
    
  center_x_scale100  = (int) (c.x()/scale);
  center_y_scale100  = (int) (c.y()/scale);
}

void LabelCanvas::moveBy(double dx, double dy) {
  QCanvasText::moveBy(dx, dy);
  if (! the_canvas()->do_zoom())
    set_center100();
}


void LabelCanvas::move_outside(QRect r, double angle) {
  QPoint center(r.center());
  QRect lbl_r(0, 0, width(), height());
  
  // place it centered on r
  lbl_r.moveCenter(center);
  
  // goes outside r, moving by 5 points
  double cx = center.x();
  double cy = center.y();
  double d_x = 5*cos(angle * 3.1415927 / 180);
  double d_y = 5*sin(angle * 3.1415927 / 180);
  
  do {
    cx += d_x;
    cy += d_y;
    
    QPoint p((int) cx, (int) cy);
    
    lbl_r.moveCenter(p);
  } while (lbl_r.intersects(r));
  
  move(lbl_r.x(), lbl_r.y());
}

void LabelCanvas::draw(QPainter & p) {
  if (! visible()) return;
  
  p.setBackgroundMode(QObject::TransparentMode);
  
  QCanvasText::draw(p);
  
  if (selected())
    show_mark(p, boundingRect());
}

UmlCode LabelCanvas::type() const {
  return UmlLabel;
}

int LabelCanvas::rtti() const {
  return RTTI_LABEL;
}

QPoint LabelCanvas::center() const {
  return boundingRect().center();
}

bool LabelCanvas::contains(int, int) const {
  return FALSE;
}

const char * LabelCanvas::may_start(UmlCode &) const {
  return "illegal";
}

const char * LabelCanvas::may_connect(UmlCode, const DiagramItem *) const {
  return "illegal";
}

void LabelCanvas::connexion(UmlCode, DiagramItem *,
			    const QPoint &, const QPoint &) {
  // no connexion available
}

void LabelCanvas::open() {
}

void LabelCanvas::menu(const QPoint&) {
}

double LabelCanvas::get_z() const {
  return z();
}

void LabelCanvas::set_z(double z) {
  setZ(z);
}

//

void LabelCanvas::save(QTextStream  & st, bool, QString &) const {
  st << "label ";
  save_string(text(), st);
  if (font().bold())
    st << " bold";
  if (font().italic())
    st << " italic";
  if (font().underline())
    st << " underlined";
  save_xyz(st, this, " xyz");
}

LabelCanvas * LabelCanvas::read(char * & st, UmlCanvas * canvas, char * k)
{
  if (strcmp(k, "label"))
    return 0;
  
  QString s = read_string(st);
  bool bold = FALSE;
  bool italic = FALSE;
  bool underlined = FALSE;
  
  for (;;) {
    k = read_keyword(st);
    
    if (!strcmp(k, "bold"))
      bold = TRUE;
    else if (!strcmp(k, "italic"))
      italic = TRUE;
    else if (!strcmp(k, "underlined"))
      underlined = TRUE;
    else
      break;
  }

  int x = (int) read_double(st);
  
  LabelCanvas * result =
    new LabelCanvas(s, canvas, x, (int) read_double(st),
		    bold, italic, underlined);
  
  if (!strcmp(k, "xyz"))
    result->setZ(read_double(st));
  else if (!strcmp(k, "xy"))
    // old version
    ;//result->setZ(OLD_LABEL_Z);
  else
    wrong_keyword(k, "xyz");
  
  return result;
}
