// *************************************************************************
//
// Copyright (C) 2004,2005 Bruno PAGES  All rights reserved.
//
// This file is part of the BOUML Uml Toolkit.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
// e-mail : bouml@free.fr
// home   : http://bouml.free.fr
//
// *************************************************************************

#ifndef UMLENUM_H
#define UMLENUM_H

#define RTTI_ARROW	1000
#define RTTI_COL_MSG	1001
#define RTTI_LABEL	1002

// warning : enums items order are linked with the ones in api.h
// and known by ToolCom.cpp

enum UmlCode {
  UmlAssociation, UmlDirectionalAssociation, UmlGeneralisation, 
  UmlAggregation, UmlAggregationByValue, UmlDirectionalAggregation,
  UmlDirectionalAggregationByValue, UmlDependency, UmlRealize,
  UmlRelations, // the first one just after the relations
  UmlAttribute, UmlOperation, UmlExtraMember,
  UmlClass,
  // all the previous ones are class members
  UmlUseCase, UmlComponent, UmlDeploymentNode, UmlArtifact,
  UmlSimpleRelations,	// mainly defined for api
  UmlClassDiagram, UmlUseCaseDiagram, UmlSeqDiagram, UmlColDiagram,
  UmlComponentDiagram, UmlDeploymentDiagram,
  UmlClassView, UmlUseCaseView, UmlComponentView, UmlDeploymentView,
  UmlPackage,
  UmlState, UmlTransition, UmlRegion, UmlStateDiagram, UmlStateAction,
  InitialPS, EntryPointPS,	// warning : PS order used
  FinalPS, TerminatePS, ExitPointPS,
  DeepHistoryPS, ShallowHistoryPS, 
  JunctionPS, ChoicePS, ForkPS, JoinPS,
  // the followings are unknown by api.h
  UmlDependOn, UmlInherit,	// not between classes
  UmlArrowPoint,
  UmlNote, UmlAnchor, UmlLabel, UmlText,
  UmlSyncMsg, UmlAsyncMsg, UmlSyncSelfMsg, UmlAsyncSelfMsg, UmlReturnMsg,
  UmlLifeLine, UmlActivityDuration,
  UmlLink, UmlSelfLink, UmlLinkDirs,
  UmlContain, 
  UmlTemplate, UmlIcon,
  UmlProject, UmlSelect,
  UmlHub,
  UmlRequired, UmlProvided, UmlArrowJunction,
  UmlFragment, UmlContinuation,
  UmlPseudoState,		// usefull in some case
  UmlSignalIn,
  UmlSignalOut,

  UmlCodeSup
};

#define IsaRelation(x) (x < UmlRelations)
#define IsaPseudoState(x) ((x >= InitialPS) && (x <= JoinPS))

enum aCorner { NoCorner, TopLeft, TopRight, BottomLeft, BottomRight };

enum UmlVisibility { UmlPublic, UmlProtected, UmlPrivate, UmlDefaultVisibility };

enum UmlParamDirection { UmlInOut, UmlIn, UmlOut };

enum Uml3States { UmlYes, UmlNo, UmlDefaultState };

enum UmlColor { UmlWhite, UmlYellow, UmlDarkYellow, UmlBlue, UmlDarkBlue, UmlGreen, UmlDarkGreen, UmlRed, UmlGray, UmlDefaultColor };

enum ClassDrawingMode { asClass, asControl, asBoundary, asEntity, asActor, Natural, DefaultClassDrawingMode };

enum DrawingLanguage { UmlView, CppView, JavaView, IdlView, DefaultDrawingLanguage };

enum ShowContextMode { noContext, umlContext, namespaceContext, packageContext, moduleContext, DefaultShowContextMode };

enum UmlFont {
  UmlSmallFont, UmlSmallItalicFont, UmlSmallBoldFont, UmlSmallBoldItalicFont, 
  UmlNormalFont, UmlNormalItalicFont, UmlNormalBoldFont, UmlNormalBoldItalicFont, 
  UmlNormalUnderlinedFont, UmlNormalStrikeOutFont,
  UmlLargeFont, UmlLargeItalicFont, UmlLargeBoldFont, UmlLargeBoldItalicFont,
  UmlFontNumber
};

enum CanvasFormat {
  IsoA2, IsoA3, IsoA4, IsoA5, UsA, UsB, UsC, CanvasFormatSup
};

extern const char * stringify(UmlVisibility);
extern const char * stringify(UmlParamDirection);
extern const char * stringify(UmlCode);	// lines only
extern const char * stringify(Uml3States);
extern const char * stringify(UmlColor);
extern const char * stringify(UmlFont);
extern const char * stringify(ClassDrawingMode);
extern const char * stringify(DrawingLanguage);
extern const char * stringify(ShowContextMode);
extern const char * stringify(CanvasFormat);
extern UmlCode arrow_type(const char *);
extern bool IsaSimpleRelation(UmlCode);
extern UmlVisibility visibility(const char * s);
extern UmlCode relation_type(const char * s,
			     bool simple_relation = FALSE);
extern UmlParamDirection direction(const char * s);
extern Uml3States state(const char * s);
extern UmlColor color(const char * s);
extern UmlFont font(const char * s);
extern ClassDrawingMode drawing_mode(const char * s);
extern DrawingLanguage drawing_language(const char * s);
extern ShowContextMode context_mode(const char * s);
extern CanvasFormat canvas_format(const char * s);
extern UmlCode pseudo_state_kind(const char * s);

#endif
